// IPC module definition.

#include "Cipc.hpp"
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include "EXTERN.h"
#include "perl.h"
#include "XSUB.h"


DWORD WaitForMultiple(CPerl* pPerl,AV* hArray,BOOL fWaitAll,DWORD dwTimeOut)
{
	int	iNumberOfElements;
	SV *	svCurrSV;
	IV		tmp;
	Cipc *	CurrObj;
	HANDLE hHandle;
	HANDLE* aHandles;
	int	i=0;
	DWORD bRet;

	iNumberOfElements = av_len( hArray );

	if( iNumberOfElements == -1 ){
		fprintf(stderr,"WaitForMultipleObjects: No Objects to wait for!\n");
		return( WAIT_FAILED );	
	}

	aHandles = new HANDLE[iNumberOfElements];

	// Create the array of handles for the WaitForMultipleObjects call
	
	for(i=0;i<iNumberOfElements+1;i++){
		svCurrSV = (SV *)(av_pop( hArray ));

	// Check if the object reference is valid

	if( sv_isa(svCurrSV,"Win32::Process") ||\
		 sv_isa(svCurrSV,"Win32::Mutex") || \
		 sv_isa(svCurrSV,"Win32::Semaphore") ||\
		 sv_isa(svCurrSV,"Win32::ChangeNotification")){

			tmp = SvIV( (SV *)SvRV( svCurrSV));
			CurrObj = (Cipc *)tmp;
		}

	else{			 
		fprintf(stderr,"WaitForMultipleObjects: Can only wait on Semaphore,mutex,ChangeNotification, or Process Objects.\n");
		return(WAIT_FAILED);
		}

// exception handling seems to be the easiest way to check if the
// object hasn't been trampled on.

		assert( CurrObj != NULL);

		try{
			hHandle = CurrObj->MyHandle();

// Check if the hHandle has been changed.

			if( hHandle == 0 ){
				fprintf(stderr,"WaitForMultipleObjects: invalid object passed\n");
				return(WAIT_FAILED);
				}
	
			aHandles[i]=hHandle;
			}
		catch(...){
			fprintf(stderr,"WaitForMultipleObjects: Exception occured while processing handles\n");
			return(WAIT_FAILED);
			}//try..catch

// Seems like we made it to here.
// So lets wait for the objects.

	}// for loop

	bRet = WaitForMultipleObjects(iNumberOfElements + 1,aHandles,fWaitAll,dwTimeOut);
	
	delete aHandles;
	
	return(bRet);


}

static double
constant(CPerl* pPerl,char* name,int arg)
{
    errno = 0;
    switch (*name) {
    case 'A':
	break;
    case 'B':
	break;
    case 'C':
	break;
    case 'D':
	break;
    case 'E':
	break;
    case 'F':
	break;
    case 'G':
	break;
    case 'H':
	break;
    case 'I':
	if (strEQ(name, "INFINITE"))
#ifdef INFINITE
	    return INFINITE;
#else
	    goto not_there;
#endif
	break;
    case 'J':
	break;
    case 'K':
	break;
    case 'L':
	break;
    case 'M':
	break;
    case 'N':
	break;
    case 'O':
	break;
    case 'P':
	break;
    case 'Q':
	break;
    case 'R':
	break;
    case 'S':
	break;
    case 'T':
	break;
    case 'U':
	break;
    case 'V':
	break;
    case 'W':
	break;
    case 'X':
	break;
    case 'Y':
	break;
    case 'Z':
	break;
    }
    errno = EINVAL;
    return 0;

not_there:
    errno = ENOENT;
    return 0;
}


XS(XS_NT__IPC_constant)
{
    dXSARGS;
    if (items != 2) {
	croak("Usage: I::constant(name,arg)");
    }
    {
	char *	name = (char *)SvPV(ST(0),na);
	int	arg = (int)SvIV(ST(1));
	double	RETVAL;

	RETVAL = constant(pPerl,name, arg);
	ST(0) = sv_newmortal();
	sv_setnv(ST(0), (double)RETVAL);
    }
    XSRETURN(1);
}


XS(XS_NT__IPC_WaitForMultiple)
{
    dXSARGS;
    if (items != 3) {
	croak("Usage: Win32::IPC::WaitForMultiple($ArrayRef,$WaitAll,$TimeOut)");
    }
    {
	SV*	hArray = ST(0);
	BOOL	fWaitAll = (BOOL)SvIV(ST(1));
	DWORD	dwTimeOut = (double)SvNV(ST(2));
	DWORD	RETVAL;
	DWORD 	BRet;

	if( ! SvROK( hArray )){
		croak("WaitForMultiple: First arg must be an array reference.");
	}

	BRet = WaitForMultiple(pPerl,(AV *)SvRV(hArray),fWaitAll,dwTimeOut);
	RETVAL = ( BRet != WAIT_FAILED )&&(BRet != WAIT_TIMEOUT);
	
	ST(0) = sv_newmortal();
	sv_setnv(ST(0), (double)RETVAL);
    }
    XSRETURN(1);
}

extern "C"
XS(boot_Win32__IPC)
{
    dXSARGS;
    char* file = __FILE__;
    newXS("Win32::IPC::constant", XS_NT__IPC_constant, file);
    newXS("Win32::IPC::WaitForMultipleObjects", XS_NT__IPC_WaitForMultiple, file);
    ST(0) = &sv_yes;
    XSRETURN(1);
}
