#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <io.h>
#include <windows.h>            // UGH!

#define SHELL_FLAG "/C"

#define REDIRECTION_STDIN           1
#define REDIRECTION_STDOUT          2
#define REDIRECTION_STDOUT_APPEND   4
#define REDIRECTION_STDERR          8
#define REDIRECTION_STDERR_APPEND  16
#define REDIRECTION_STDERR_STDOUT  32
#define REDIRECTION_STDERR_NULL    64

// Function allusions
static int SpawnWait(char *cmdLine, char *cpChildEnv);
unsigned int HasRedirection(char* ptr,
							char* inputFilename,
							char* outputFilename,
							char* errorFilename);

// ==========
// MAIN ENTRY
// ==========
int main(int argc, char *argv[])
{
	int i, j;
	int iStat = 0;
	int err;
	unsigned int nRedirections;
	char szTmpFilename[512];
	char szCmd[2048];
	char szBuf[8192];
	char szCmdFile[512];
	char szInpFile[512];
	char szOutFile[512];
	char szErrFile[512];
	FILE *tcmd, *tout, *terr;

	// Nothing to do
	if(argc < 2)
		return 0;

	//
	// Make the tempfile names for the start/input/output files
	// Use a 100Hz counter to make unique name.     This wraps around
	// every 46 hours... Use 8.3 names to minimize the filesystem
	// thrashing from the House That Jack Built FAT "long filename
	// support" in Win95.
	//
	sprintf(szTmpFilename,
			"%s\\pl%06lx",
			getenv("TEMP"),
			((GetTickCount() >> 2) & 0x00FFFFFF));
	sprintf(szCmdFile, "%s.bat", szTmpFilename);
	sprintf(szErrFile, "%s.err", szTmpFilename);
	szInpFile[0] = szOutFile[0] = '\0';

	//
	// Now create the command file to feed to the brain-dead
	// command interpreter.
	//
	if((tcmd = fopen(szCmdFile, "wt")) == NULL)
	{
		fprintf(stderr, "Error creating cmd file (%d)\n", errno);
		iStat = -1;
		goto cleanup;
	}

	//
	// Reconstruct the command line from the argvec
	// skip the first "/c" if present; will be added later
	//
	szCmd[0] = '\0';
	i = 1;
	if(stricmp(argv[i], SHELL_FLAG) == 0)
		++i;

	while(i < argc)
	{
		strcat(szCmd, argv[i++]);
		strcat(szCmd, " ");
	}
	szCmd[strlen(szCmd) - 1] = '\0';        // Zap tailing blank

#if defined DEBUG
	printf("New command line: %s!\n", szCmd);
#endif

	nRedirections = HasRedirection(szCmd, szInpFile, szOutFile, szErrFile);

#if defined DEBUG
	if(nRedirections & REDIRECTION_STDIN)
		printf("REDIRECTION_STDIN\n");
	if(nRedirections & REDIRECTION_STDOUT)
		printf("REDIRECTION_STDOUT\n");
	if(nRedirections & REDIRECTION_STDOUT_APPEND)
		printf("REDIRECTION_STDOUT_APPEND\n");
	if(nRedirections & REDIRECTION_STDERR)
		printf("REDIRECTION_STDERR\n");
	if(nRedirections & REDIRECTION_STDERR_APPEND)
		printf("REDIRECTION_STDERR_APPEND\n");
	if(nRedirections & REDIRECTION_STDERR_STDOUT)
		printf("REDIRECTION_STDERR_STDOUT\n");
#endif

	if(nRedirections)
	{
		if(nRedirections & REDIRECTION_STDERR)
		{
			if((nRedirections & REDIRECTION_STDOUT) && !strcmp(szErrFile, szOutFile))
			{
				nRedirections |= REDIRECTION_STDERR_STDOUT;
				sprintf(szErrFile, "%s.err", szTmpFilename);
			}
			else
			{
				if(!strcmp(szErrFile, "NUL"))
				{
					nRedirections |= REDIRECTION_STDERR_NULL;
					sprintf(szErrFile, "%s.err", szTmpFilename);
				}
			}
		}
	}
#if defined DEBUG
	printf("szInpFile = %s\n", szInpFile);
	printf("szOutFile = %s\n", szOutFile);
	printf("szErrFile = %s\n", szErrFile);
#endif

	// Fix up any %'s in the command, since DOS munges them in the batch file
	for(i = 0, j = 0; (szCmd[i] != '\0') && (i < 2048); i++, j++)
	{
		szBuf[j] = szCmd[i];
		if(szCmd[i] == '%')
		{
			szBuf[++j] = '%';
		}
	}
	if(i < 2048)
		szBuf[j] = '\0';
	else
		szBuf[2047] = '\0';

	memcpy( szCmd, szBuf, 2048 );

	//
	// Write out the command to the jacket proc
	//
	fprintf(tcmd, "@ECHO OFF\n");

	if(nRedirections & REDIRECTION_STDIN)
	{
		sprintf(szBuf, " < %s", szInpFile);
		strcat(szCmd, szBuf);
	}

	if(nRedirections & REDIRECTION_STDOUT)
	{
		if(nRedirections & REDIRECTION_STDOUT_APPEND)
		{
			sprintf(szBuf, " >> %s", szOutFile);
		}
		else
		{
			sprintf(szBuf, " > %s", szOutFile);
		}
		strcat(szCmd, szBuf);
	}
	fprintf(tcmd, "%s\n", szCmd);

	fclose(tcmd);

	if(nRedirections & REDIRECTION_STDERR)
	{
		if(nRedirections & REDIRECTION_STDERR_APPEND)
		{
			if((terr = freopen(szErrFile, "a", stderr)) == NULL)
			{
				fprintf(stderr, "Error opening error output file %s\n", szErrFile);
			}
			fseek(terr, 0, SEEK_END);
		}
		else
		{
			if((terr = freopen(szErrFile, "w", stderr)) == NULL)
			{
				fprintf(stderr, "Error creating error output file %s\n", szErrFile);
			}
		}
	}

	//
	// Launch the subshell and wait till it's finished
	//
	sprintf(szBuf, "%s %s %s", getenv("COMSPEC"), SHELL_FLAG, szCmdFile);
	if((i = SpawnWait(szBuf, NULL)) != 0)
	{
		iStat = i;
		goto cleanup;
	}

	if(nRedirections & REDIRECTION_STDERR)
	{
		fclose(terr);
	}

	if(nRedirections & REDIRECTION_STDERR_STDOUT)
	{
		if((terr = fopen(szErrFile, "rb")) != NULL)
		{
			if((tout = fopen(szOutFile, "ab")) != NULL)
			{
				fseek(tout, 0, SEEK_END);
				err = FALSE;
				while(!feof(terr))
				{
					i = fread(szBuf, 1, 8192, terr);
					if(i == 0)
					break;
					fwrite(szBuf, 1, i, tout);

					if(ferror(tout))
					{
						err = TRUE;
						break;
					}
				}
				fclose(tout);
			}

			if(err)
			{
				fprintf(stderr, "Error copying tempfile %s to %s\n", szErrFile, szOutFile);
				iStat = -1;
			}
			fclose(terr);
		}
	}

cleanup:

#if !defined DEBUG
	remove(szCmdFile);

	if((nRedirections & REDIRECTION_STDERR_STDOUT) ||
		(nRedirections & REDIRECTION_STDERR_NULL))
	{
		remove(szErrFile);
	}
#endif
	return(iStat);
}


//
// HasRedirection
//
unsigned int HasRedirection(char* ptr,
							char* inputFilename,
							char* outputFilename,
							char* errorFilename)
{
	int  inquote = 0;
	char quote = '\0';
	int  redirectionFound = 0;
	char chRedirectType;
	int  i;

	//
	// Scan string looking for redirection (< or >) or pipe
	// characters (|) that are not in a quoted string
	//
	while(*ptr)
	{
		switch(*ptr)
		{
			case '\'':
			case '\"':
				if(inquote)
				{
					if(quote == *ptr)
					{
						inquote = 0;
						quote = '\0';
					}
				}
				else
				{
					quote = *ptr;
					inquote++;
				}
				break;

			case '>':
			case '<':
				if(!inquote)
				{

					if(*ptr == '<')
					{
						/* Input redirection */
						redirectionFound |= REDIRECTION_STDIN;

						*ptr++ = '\0';

						while(*ptr == ' ')
							ptr++;

						i = 0;
						while(*ptr != ' ' && *ptr != '\0')
							inputFilename[i++] = *ptr++;

						inputFilename[i] = '\0';
					}
					else
					{
						/* Output redirection */
						chRedirectType = *(ptr - 1);

						switch(chRedirectType)
						{
							case '1':
							case '2':
								if(*(ptr - 2) != ' ')
								{
									/* The 1 or 2 is part of the command, our redirection */
									/* is of stdout.                                      */
									/* e.g. "foo1>out"                                    */
									*ptr++ = '\0';
									chRedirectType = '1';
								}
								else
								{
									/* The 1 or 2 specifies which redirection to use.     */
									/* e.g. "foo 1>out" or "foo 2>out"                    */
									*(ptr++ - 2) = '\0';
								}
								break;

							case ' ':
								/* Our redirection is of stdout.                         */
								/* e.g. "foo >out"                                       */
								*(ptr++ - 1) = '\0';
								chRedirectType = '1';
								break;

							default:
								/* Our redirection is of stdout.                         */
								/* e.g. "foo>out"                                        */
								*ptr++ = '\0';
								chRedirectType = '1';
								break;
						}

						/* Check if the redirection is being appended */
						if(*ptr == '>')
						{
							if(chRedirectType == '1')
							{
								redirectionFound |= REDIRECTION_STDOUT_APPEND;
							}
							else
							{
								redirectionFound |= REDIRECTION_STDERR_APPEND;
							}
							ptr++;
						}

						if(chRedirectType == '2')
						{
							/* Standard error redirection, get filename */
							redirectionFound |= REDIRECTION_STDERR;

							if((ptr[0] == '&') && (ptr[1] == '1'))
							{
								redirectionFound |= REDIRECTION_STDERR_STDOUT;
								ptr += 2;
							}
							else
							{
								while(*ptr == ' ')
								ptr++;

								i = 0;
								while(*ptr != ' ' && *ptr != '\0')
								{
									if(*ptr == '/')
									{
										errorFilename[i++] = '\\';
										ptr++;
									}
									else
									{
										errorFilename[i++] = *ptr++;
									}
								}
								errorFilename[i] = '\0';
							}
						}
						else
						{
							/* Standard output redirection */
							redirectionFound |= REDIRECTION_STDOUT;

							while(*ptr == ' ')
								*ptr++;

							i = 0;
							while(*ptr != ' ' && *ptr != '\0')
							{
								if(*ptr == '/')
								{
									outputFilename[i++] = '\\';
									ptr++;
								}
								else
								{
									outputFilename[i++] = *ptr++;
								}
							}
								outputFilename[i] = '\0';
						}
					}
				}

			default:
				break;
		}
		++ptr;
	}
	return redirectionFound;
}



//------------------------------------------------------------------------
//
// SpawnWait() - Spawn a child process, wait for termination
//
// Use CreateProcess() to fire off a child process, then synchronize with
// its termination. Child window is hidden unless being debugged.
//
// Inputs:
//    cmdLine     Command line including executable name
//    envPtr      Pointer to Win32 environment block for child (NULL os OK).
//
// Returns:
//    -1          Big Trouble: WaitForSingleObject (child) failed or
//                                failed to get exit status of child
//    else        Exit status of child
//
// On Error:
//
//------------------------------------------------------------------------
static int SpawnWait(char *cmdLine, char *cpChildEnv)
{
	PROCESS_INFORMATION stProcInfo;
	STARTUPINFO stStartInfo;
	DWORD dwChildExitStat;

	//
	// Create the child process
	//**TODO - Security? Inherited handles? Can this become like the Unix
	// CGI with environment variables and inherited socket handles? Also,
	// variable priority? Should it be lower priority?
	//
	memset(&stStartInfo, 0, sizeof(stStartInfo));   // Clear the block
	stStartInfo.cb = sizeof(stStartInfo);           // Set the structure size
	stStartInfo.dwFlags = STARTF_USESHOWWINDOW; // Enable wShowWindow control
	stStartInfo.wShowWindow = SW_SHOWMINNOACTIVE;   // Start min (normal)
	//
	// This is it...
	//
	if(!CreateProcess(NULL,                   // Use command line for image
					cmdLine,                // Has everything including image path
					NULL,                   // Default process security
					NULL,                   // Default thread security
					FALSE,                  // Must be TRUE to use std handles
					NORMAL_PRIORITY_CLASS,  // No special scheduling
					(LPVOID)cpChildEnv,     // Environment block or NULL
					NULL,                   // Inherit our currrent directory
					&stStartInfo,           // -> Startup info
					&stProcInfo))           // <- Process info (if OK)
	{
		fprintf(stderr,
				"SpawnWait: CreateProcess() failed. Code = %ld\n",
					GetLastError());
		return(-1);
	}

	//
	// Now simply hang around until the thread exits
	//**TODO: Timer here?
	//
	if(WaitForSingleObject(stProcInfo.hProcess, INFINITE) != WAIT_OBJECT_0)
	{
		fprintf(stderr, "Child process sync failed.\n");
		dwChildExitStat = 0xFFFFFFFF;           // Big trouble
	}
	else if(!GetExitCodeProcess(stProcInfo.hProcess, &dwChildExitStat))
	{
		fprintf(stderr, "Failed to get child exit status\n");
		dwChildExitStat = 0xFFFFFFFF;   //TODO - print?
	}
	CloseHandle(stProcInfo.hProcess);
	CloseHandle(stProcInfo.hThread);

	return(dwChildExitStat);
}
