//	IPerlSup.h
//	Support interface for perl

#ifndef __Inc__IPerlSupport___
#define __Inc__IPerlSupport___

#ifndef HINSTANCE
typedef void *HINSTANCE;
#endif

#ifndef HANDLE
typedef void *HANDLE;
#endif

class IPerlSupport
{
public:
	// memory allocation routines
	virtual void* Malloc(size_t) = 0;
	virtual void* Realloc(void*, size_t) = 0;
	virtual void Free(void*) = 0;

	// FILE* routines
	virtual FILE* Stdin(void) = 0;
	virtual FILE* Stdout(void) = 0;
	virtual FILE* Stderr(void) = 0;
	virtual void Clearerr(FILE *) = 0;
	virtual int Fclose(FILE *, int&) = 0;
	virtual FILE* Fdopen(int, const char *, int&) = 0;
	virtual int Feof(FILE *, int&) = 0;
	virtual int Ferror(FILE *, int&) = 0;
	virtual int FFlush(FILE *, int&) = 0;
	virtual int Flushall(int&) = 0;
	virtual int Fgetc(FILE *, int&) = 0;
	virtual int FGetpos(FILE *, fpos_t *, int&) = 0;
	virtual char* Fgets(char *, int, FILE *, int&) = 0;
	virtual int Fileno(FILE *, int&) = 0;
	virtual FILE* Fopen(const char *, const char *, int&) = 0;
	virtual int Fputs(const char *, FILE *, int&) = 0;
	virtual size_t Fread(void *, size_t, size_t, FILE *, int&) = 0;
	virtual int Fseek(FILE *, long, int, int&) = 0;
	virtual int FSetpos(FILE *, fpos_t *, int&) = 0;
	virtual int FStat(int, struct stat *, int&) = 0;
	virtual long Ftell(FILE *, int&) = 0;
	virtual size_t Fwrite(const void *, size_t, size_t, FILE *, int&) = 0;
	virtual int Getc(FILE *, int&) = 0;
	virtual int Putc(int c, FILE *, int&) = 0;
	virtual void Rewind(FILE *) = 0;
	virtual void Setbuf(FILE *, char *, int&) = 0;
	virtual int Setvbuf(FILE *, char *, int, size_t, int&) = 0;
	virtual FILE* TmpFile(int&) = 0;
	virtual int Ungetc(int, FILE *, int&) = 0;
	virtual int Vfprintf(FILE *, const char *, va_list, int&) = 0;
	virtual FILE* Popen(const char *, const char *, int&) = 0;
	virtual int Pclose(FILE *, int&) = 0;
	virtual int Pipe(int *, unsigned int, int, int&) = 0;

	// Low level I/O
	virtual int Access(const char *, int, int&) = 0;
	virtual int Chmod(const char *, int, int&) = 0;
	virtual int Chsize(int, long, int&) = 0;
	virtual int Close(int, int&) = 0;
	virtual int Dup(int, int&) = 0;
	virtual int Dup2(int, int, int&) = 0;
	virtual int Isatty(int, int&) = 0;
	virtual long Lseek(int, long, int, int&) = 0;
	virtual char *Mktemp(char *, int&) = 0;
	virtual int Open(const char *, int, int, int&) = 0;
	virtual int Read(int, void *, unsigned int, int&) = 0;
	virtual int Rename(const char *, const char *, int&) = 0;
	virtual int Setmode(int, int, int&) = 0;
	virtual char *Tmpnam(char *, int&) = 0;
	virtual int Umask(int, int&) = 0;
	virtual int Unlink(const char *, int&) = 0;
	virtual int Utime(const char *filename, struct utimbuf *times, int&) = 0;
	virtual int Write(int, const void *, unsigned int, int&) = 0;

	// Spawning & Exec'ing
	virtual int Spawn(int mode, const char *cmdname, const char * const *argv, const char * const *envp, int&) = 0;
	virtual int Cmd(int mode, const char *shell, const char *cmd, const char * const *envp) = 0;
	virtual int Execv(const char *cmdname, const char *const *argv) = 0;
	virtual int Execvp(const char *cmdname, const char *const *argv) = 0;

	// Sockets
	virtual int OpenOSfhandle(long osfhandle, int flags) = 0;
	virtual int GetOSfhandle(int filenum) = 0;

	// Misc
	virtual int SetSignal(int i) = 0;
	virtual void SignalHandler(CPerl *pPerl, int sig) = 0;
	virtual void Die(CPerl *pPerl, char *message) = 0;
	virtual void Putenv(const char *envstring) = 0;
	virtual char* Getenv(const char *envstring) = 0;
	virtual int GetCWD(char *buffer, int size) = 0;
	virtual int SetCWD(char *buffer) = 0;
	virtual void PerlCaller(char *string, int size) = 0;
	virtual void* GetECB(void) = 0;

	virtual int MkDir(char *, int&) = 0;
	virtual int ChDir(char *, int&) = 0;
	virtual int RmDir(char *, int&) = 0;
	virtual int FileStat(char *, void *, int&) = 0;
	virtual char *GetStrerror(int) = 0;

	virtual void XInit(void) = 0;

	// find file functions
	virtual HANDLE FindFirstFile(const char *, void*) = 0;
	virtual int FindNextFile(HANDLE, void *) = 0;
	virtual int FindClose(HANDLE) = 0;
};

#endif	// __Inc__IPerlSupport___

