// NTLib.c

// (c) 1995 Microsoft Corporation. All rights reserved. 
//	Developed by ActiveWare Internet Corp., http://www.ActiveWare.com

//    You may distribute under the terms of either the GNU General Public
//    License or the Artistic License, as specified in the README file.



#include <windows.h>
#define WIN32_LEAN_AND_MEAN

#include "EXTERN.h"
#include "perl.h"

#include "XSub.h"

#define CROAK croak
#define newXSInternal    pPerl->Perl_newXS

#define TMPBUFSZ 512

#define SUCCESSRETURNED(x)	(x == ERROR_SUCCESS)

#define SETIV(index,value) sv_setiv(ST(index), value)
#define SETPV(index,string) sv_setpv(ST(index), string)
#define SETPVN(index, buffer, length) sv_setpvn(ST(index), (char*)buffer, length)
#define PERLSvIV(sv) (SvIOK(sv) ? SvIVX(sv) : sv_2iv(sv))
#define PERLSvPV(sv, lp) (SvPOK(sv) ? ((lp = SvCUR(sv)), SvPVX(sv)) : sv_2pv(sv, &lp))

#define RETURNRESULT if(bSuccess){ XST_mYES(0); }\
					 else {	XST_mNO(0); }\
					 XSRETURN(1)

static char thisFile[] = __FILE__;


XS(NTLoadLibrary)
{
	dXSARGS;
	unsigned int length;

	if(items != 1) 
	{
		CROAK ("usage: Win32::LoadLibrary($LibraryName)\n");
	}

	XSRETURN_IV((long)LoadLibrary((char *) PERLSvPV(ST(0), length)));
}

XS(NTFreeLibrary)
{
	dXSARGS;
	int length;

	if(items != 1) 
	{
		CROAK ("usage: Win32::FreeLibrary($hinstance)\n");
	}

	if(FreeLibrary((HINSTANCE) PERLSvIV(ST(0))))
	{
		XSRETURN_YES;
	}
	XSRETURN_NO;
}

XS(NTGetProcAddress)
{
	dXSARGS;
	unsigned int length;

	if(items != 2) 
	{
		CROAK ("usage: Win32::GetProcAddress($hinstance, $procedurename)\n");
	}

	XSRETURN_IV((long)GetProcAddress((HINSTANCE) PERLSvIV(ST(0)), (char *) PERLSvPV(ST(1), length)));
}

XS(NTRegisterServer)
{
	dXSARGS;
	BOOL bSuccess = FALSE;
	HINSTANCE hInstance;
	unsigned int length;
	FARPROC sFunc;

	if(items != 1) 
	{
		CROAK ("usage: Win32::RegisterServer($LibraryName)\n");
	}

	hInstance = LoadLibrary((char *)PERLSvPV(ST(0), length));
	if(hInstance != NULL)
	{
		sFunc = GetProcAddress(hInstance, "DllRegisterServer");
		if(sFunc != NULL)
		{
			bSuccess = (sFunc() == 0);
		}
		FreeLibrary(hInstance);
	}

	RETURNRESULT;
}

XS(NTUnregisterServer)
{
	dXSARGS;
	BOOL bSuccess = FALSE;
	HINSTANCE hInstance;
	unsigned int length;
	FARPROC sFunc;

	if(items != 1) 
	{
		CROAK ("usage: Win32::UnregisterServer($LibraryName)\n");
	}

	hInstance = LoadLibrary((char *)PERLSvPV(ST(0), length));
	if(hInstance != NULL)
	{
		sFunc = GetProcAddress(hInstance, "DllUnregisterServer");
		if(sFunc != NULL)
		{
			bSuccess = (sFunc() == 0);
		}
		FreeLibrary(hInstance);
	}

	RETURNRESULT;
}


void NTLibInit(CPerl *pPerl)
{
	newXSInternal("Win32::LoadLibrary", NTLoadLibrary, thisFile);
	newXSInternal("Win32::FreeLibrary", NTFreeLibrary, thisFile);
	newXSInternal("Win32::GetProcAddress", NTGetProcAddress, thisFile);
	newXSInternal("Win32::RegisterServer", NTRegisterServer, thisFile);
	newXSInternal("Win32::UnregisterServer", NTUnregisterServer, thisFile);
}

