// NTOLE.cpp

// (c) 1995 Microsoft Corporation. All rights reserved. 
//	Developed by ActiveWare Internet Corp., http://www.ActiveWare.com

//    You may distribute under the terms of either the GNU General Public
//    License or the Artistic License, as specified in the README file.

#include <windows.h>
#include <objbase.h>
#include <winnt.h>
#include <malloc.h>
#define WIN32_LEAN_AND_MEAN

#pragma  warning(disable: 4091)

#include "EXTERN.h"
#include "perl.h"

#include "XSub.h"



#ifdef _DEBUG
inline void ODS(LPSTR x)
{
	OutputDebugString(x);
	OutputDebugString("\n");
}

#if defined(UNICODE)
#define ODSOLE(x) ODS(x)
#else
inline void ODSOLE(LPCWSTR x)
{
	char bufA[256];

	if(x != NULL)
		WideCharToMultiByte(CP_ACP, NULL, x, -1, bufA, 256, NULL, NULL);
	else
		strcpy(bufA, "<null>");

	ODS(bufA);
}
#endif	// UNICODE

#else
#define ODS(x)
#define ODSOLE(x)
#endif	// _DEBUG

#define CROAK croak
#define newXSInternal    pPerl->Perl_newXS
#define SUCCESSRETURNED(x)	(x == ERROR_SUCCESS)
#define SETIV(index,value) sv_setiv(ST(index), value)
#define SETSV(index,value) sv_setsv(ST(index), value)
#define SETNV(index,value) sv_setnv(ST(index), value)
#define SETPV(index,string) sv_setpv(ST(index), string)
#define SETPVN(index, buffer, length) sv_setpvn(ST(index), (char*)buffer, length)
#define RETURNRESULT if(bSuccess){ XST_mYES(0); }\
					 else {	XST_mNO(0); }\
					 XSRETURN(1)

#define NEW(x,v,n,t)  (v = (t*)safemalloc((MEM_SIZE)((n) * sizeof(t))))

static char thisFile[] = __FILE__;
const DWORD WINOLE_MAGIC = 0x12344321;
const int bufferSize = 1024;

const LCID lcidDefault = (0x02 << 10) /* LOCALE_SYSTEM_DEFAULT */;

#define OLEAPI STDAPICALLTYPE

static HINSTANCE hOleLib = NULL;
static HINSTANCE hOleAutoLib = NULL;
extern CRITICAL_SECTION csOLE;

// functions from ole32.dll
typedef HRESULT (OLEAPI *LPOleInitialize)(LPVOID);
typedef void (OLEAPI *LPOleUninitialize)(void);
typedef HRESULT (OLEAPI *LPCoCreateInstance)(REFCLSID, LPUNKNOWN, DWORD, REFIID, LPVOID *);
typedef HRESULT (OLEAPI *LPCLSIDFromProgID)(LPCOLESTR, LPCLSID);

static LPOleInitialize pOleInitialize = 0;
static LPOleUninitialize pOleUninitialize = 0;
static LPCoCreateInstance pCoCreateInstance = 0;
static LPCLSIDFromProgID pCLSIDFromProgID = 0;


#define OLEAUTAPI STDAPICALLTYPE

// functions from oleaut32.dll
typedef BSTR (OLEAUTAPI *LPSysAllocString)(OLECHAR* sz);
typedef void (OLEAUTAPI *LPSysFreeString)(BSTR);
typedef unsigned int (OLEAUTAPI *LPSysStringLen)(BSTR);
typedef HRESULT (OLEAUTAPI *LPVariantClear)(VARIANTARG *);
typedef void (OLEAUTAPI *LPVariantInit)(VARIANTARG *);
typedef SAFEARRAY* (OLEAUTAPI *LPSafeArrayCreate)(VARTYPE, unsigned int, SAFEARRAYBOUND *);
typedef HRESULT (OLEAUTAPI *LPSafeArrayPutElement)(SAFEARRAY *, long *, void *);
typedef HRESULT (OLEAUTAPI *LPSafeArrayGetLBound)(SAFEARRAY *, unsigned int, long *);
typedef HRESULT (OLEAUTAPI *LPSafeArrayGetUBound)(SAFEARRAY *, unsigned int, long *);
typedef HRESULT (OLEAUTAPI *LPSafeArrayGetElement)(SAFEARRAY *, long *, void *);
typedef HRESULT (OLEAUTAPI *LPSafeArrayDestroy)(SAFEARRAY *);
typedef unsigned int (OLEAUTAPI *LPSafeArrayGetDim)(SAFEARRAY *);
typedef HRESULT (OLEAUTAPI *LPVariantChangeType)(VARIANTARG *, VARIANTARG*, unsigned short, VARTYPE);
typedef HRESULT (OLEAUTAPI *LPSafeArrayAccessData)(SAFEARRAY *, void **);
typedef HRESULT (OLEAUTAPI *LPSafeArrayUnaccessData)(SAFEARRAY *);


static LPSysAllocString pSysAllocString = 0;
static LPSysFreeString pSysFreeString = 0;
static LPSysStringLen pSysStringLen = 0;
static LPVariantClear pVariantClear = 0;
static LPVariantInit pVariantInit = 0;
static LPSafeArrayCreate pSafeArrayCreate = 0;
static LPSafeArrayPutElement pSafeArrayPutElement = 0;
static LPSafeArrayGetLBound pSafeArrayGetLBound = 0;
static LPSafeArrayGetUBound pSafeArrayGetUBound = 0;
static LPSafeArrayGetElement pSafeArrayGetElement = 0;
static LPSafeArrayDestroy pSafeArrayDestroy = 0;
static LPSafeArrayGetDim pSafeArrayGetDim = 0;
static LPVariantChangeType pVariantChangeType = 0;
static LPSafeArrayAccessData pSafeArrayAccessData = 0;
static LPSafeArrayUnaccessData pSafeArrayUnaccessData = 0; 

typedef struct _tagWINOLEOBJECT *LPWINOLEOBJECT;
typedef struct _tagWINOLEOBJECT
{
	long Win32OLEMagic;
	LPWINOLEOBJECT pNext;
	LPWINOLEOBJECT pPrevious;
	IDispatch*	pDispatch;
	IEnumVARIANT *pEnum;
	HV*	hashTable;
} WINOLEOBJECT; 

#define INTERNAL_ERROR ((HRESULT)-1)

#ifdef _DEBUG
inline void EXCEPTIONINFOCLEAR(EXCEPINFO &excepinfo)
{
	memset(&excepinfo, 0, sizeof(EXCEPINFO));
}

inline void EXCEPTIONINFO(HRESULT lastError, EXCEPINFO &excepinfo)
{
	if(FAILED(lastError))
	{
		OutputDebugString("Exception Source: ");
		ODSOLE(excepinfo.bstrSource);
		OutputDebugString("	Description: ");
		ODSOLE(excepinfo.bstrDescription);
	}
}
#else
#define EXCEPTIONINFOCLEAR(x)
#define EXCEPTIONINFO(x, y)
#endif

void ReleaseObjects(CPerl *pPerl, LPWINOLEOBJECT lpObj)
{
	if(lpObj->pDispatch != NULL)
		lpObj->pDispatch->Release();

	if(lpObj->hashTable != NULL && pPerl != NULL)
		hv_undef(lpObj->hashTable);
}

LPWINOLEOBJECT NewDispatch(CPerl *pPerl, IDispatch* pDisp, BOOL bCreated)
{
	LPWINOLEOBJECT lpObj;
	NEW(2101, lpObj, 1, WINOLEOBJECT);
	lpObj->Win32OLEMagic = WINOLE_MAGIC;
	lpObj->pNext = NULL;
	lpObj->pPrevious = NULL;
	lpObj->pEnum = NULL;
	lpObj->pDispatch = pDisp;
	lpObj->hashTable = newHV();
	lpObj->pNext = (LPWINOLEOBJECT)g_lpObj;
	g_lpObj = lpObj;

	if(!bCreated)
	{
		pDisp->AddRef();
	}

	return lpObj;
}

void ReleaseOLEObjects(CPerl *pPerl)
{	// release all unreleased OLE Objects
	LPWINOLEOBJECT lpObj;
	for(lpObj = (LPWINOLEOBJECT)g_lpObj; lpObj != NULL; lpObj = lpObj->pNext)
		ReleaseObjects(pPerl, lpObj);

	if(bOleInit)
		pOleUninitialize();
}

void __cdecl Win32OLECleanUp(void)
{
	if(hOleAutoLib != NULL)
	{
		FreeLibrary(hOleAutoLib);
		hOleAutoLib = NULL;
	}

	if(hOleLib != NULL)
	{
		FreeLibrary(hOleLib);
		hOleLib = NULL;
	}
}

static FARPROC GetOleAddress(CPerl *pPerl, HINSTANCE hInstance, char *lpFunctionName, char *lpDllName)
{
	char buffer[512];
	FARPROC proc = GetProcAddress(hInstance, lpFunctionName);
	if(proc == 0)
	{
		sprintf(buffer, "Unable to get address of %s in %s.dll", lpFunctionName, lpDllName);
		CROAK(buffer);
	}
	return proc;
}


void OleStartup(CPerl *pPerl)
{
	EnterCriticalSection(&csOLE);
	if(hOleAutoLib == NULL)
	{
		HINSTANCE hLib = LoadLibrary("OLEAut32.DLL");
		if(hLib == NULL)
			CROAK("Could not load OLEAUT32.dll\n");

		hOleLib = LoadLibrary("OLE32.DLL");
		if(hOleLib == NULL)
			CROAK("Could not load OLE32.dll\n");

		pOleInitialize = (LPOleInitialize)GetOleAddress(pPerl, hOleLib, "OleInitialize", "OLE32.DLL");
		pOleUninitialize = (LPOleUninitialize)GetOleAddress(pPerl, hOleLib, "OleUninitialize", "OLE32.DLL");;
		pCoCreateInstance = (LPCoCreateInstance)GetOleAddress(pPerl, hOleLib, "CoCreateInstance", "OLE32.DLL");
		pCLSIDFromProgID = (LPCLSIDFromProgID)GetOleAddress(pPerl, hOleLib, "CLSIDFromProgID", "OLE32.DLL");

		pSysAllocString = (LPSysAllocString)GetOleAddress(pPerl, hLib, "SysAllocString", "OLEAut32.DLL");
		pSysFreeString = (LPSysFreeString)GetOleAddress(pPerl, hLib, "SysFreeString", "OLEAut32.DLL");
		pSysStringLen = (LPSysStringLen)GetOleAddress(pPerl, hLib, "SysStringLen", "OLEAut32.DLL");
		pVariantClear = (LPVariantClear)GetOleAddress(pPerl, hLib, "VariantClear", "OLEAut32.DLL");
		pVariantInit = (LPVariantInit)GetOleAddress(pPerl, hLib, "VariantInit", "OLEAut32.DLL");
		pSafeArrayCreate = (LPSafeArrayCreate)GetOleAddress(pPerl, hLib, "SafeArrayCreate", "OLEAut32.DLL");
		pSafeArrayPutElement = (LPSafeArrayPutElement)GetOleAddress(pPerl, hLib, "SafeArrayPutElement", "OLEAut32.DLL");
		pSafeArrayGetLBound = (LPSafeArrayGetLBound)GetOleAddress(pPerl, hLib, "SafeArrayGetLBound", "OLEAut32.DLL");
		pSafeArrayGetUBound = (LPSafeArrayGetUBound)GetOleAddress(pPerl, hLib, "SafeArrayGetUBound", "OLEAut32.DLL");
		pSafeArrayGetElement = (LPSafeArrayGetElement)GetOleAddress(pPerl, hLib, "SafeArrayGetElement", "OLEAut32.DLL");
		pSafeArrayDestroy = (LPSafeArrayDestroy)GetOleAddress(pPerl, hLib, "SafeArrayDestroy", "OLEAut32.DLL");
		pSafeArrayGetDim = (LPSafeArrayGetDim)GetOleAddress(pPerl, hLib, "SafeArrayGetDim", "OLEAut32.DLL");
		pVariantChangeType = (LPVariantChangeType)GetOleAddress(pPerl, hLib, "VariantChangeType", "OLEAut32.DLL");
		pSafeArrayAccessData = (LPSafeArrayAccessData)GetOleAddress(pPerl, hLib, "SafeArrayAccessData", "OLEAut32.DLL");
		pSafeArrayUnaccessData = (LPSafeArrayUnaccessData)GetOleAddress(pPerl, hLib, "SafeArrayUnaccessData", "OLEAut32.DLL");

		hOleAutoLib = hLib;
	}
	LeaveCriticalSection(&csOLE);
}

inline void OleStartupInit(CPerl *pPerl)
{
	if(hOleAutoLib == 0)
 		OleStartup(pPerl);

	if(!bOleInit)
	{
		pOleInitialize(NULL);
		bOleInit = TRUE;
	}
}

static char *OleObjectID = "___Perl___OLEObject___";
static const int OleObjectIDLength = strlen(OleObjectID)+1;

SV *CreatePerlObject(CPerl *pPerl, IDispatch *pDisp, BOOL bCreated = FALSE)
{
	HV *hvinner, *hvouter;
	SV *blessed;

	if(pDisp != NULL)
	{
		hvinner = newHV();
		hv_store(hvinner, OleObjectID, OleObjectIDLength, newSViv((long)NewDispatch(pPerl, pDisp, bCreated)), 0);
		blessed = sv_bless(newRV((SV*)hvinner), gv_stashpv("OLEhash", TRUE));

		hvouter = newHV();
		sv_magic((SV*)hvouter, blessed, 'P', Nullch, 0);
		return sv_bless(newRV((SV*)hvouter), gv_stashpv("OLE", TRUE));
	}
	return &sv_undef;
}

#undef CreatePerlOLEObject
SV* CPerl::CreatePerlOLEObject(long pDispatch)
{
	return CreatePerlObject(this, (IDispatch*)pDispatch);
}

LPWINOLEOBJECT GetOLEObject(CPerl *pPerl, SV *sv) 
{
	SV **psv;

	if(sv != NULL)
	{
		psv = hv_fetch((HV*)SvRV(sv), OleObjectID, OleObjectIDLength, 0);
		if(psv != NULL)
			return (LPWINOLEOBJECT)SvIV(*psv);
	}
	return NULL;
}

inline BOOL IsOleStruct(LPWINOLEOBJECT lpObj)
{
	return (lpObj != NULL && lpObj->Win32OLEMagic == WINOLE_MAGIC);
}

inline BOOL ValidDispatch(LPWINOLEOBJECT lpObj)
{
	return (IsOleStruct(lpObj) && lpObj->pDispatch != NULL);
}

BSTR AllocOLEString(char* lpStr, int length)
{
	int count = (length+1)*2;
	OLECHAR* pOLEChar = (OLECHAR*)alloca(count);
	MultiByteToWideChar(CP_ACP, 0, lpStr, -1, pOLEChar, count);
	return pSysAllocString(pOLEChar);
}

XS(Win32OLELastError)
{
	dXSARGS;
	XSRETURN_IV(LastOLEError);
}

XS(Win32OLECreateObject)
{
	dXSARGS;
	CLSID CLSIDObj;
	OLECHAR bBuffer[bufferSize];
	unsigned int length;
	char *buffer;
	HKEY handle;
	IDispatch *pDisp;
	BOOL bSuccess = FALSE;

	if(items != 2)
	{
		CROAK("usage: Win32::OLECreateObject($class, $object);\n");
	}
	OleStartupInit(pPerl);

	buffer = SvPV(ST(0), length);
	MultiByteToWideChar(CP_ACP, NULL, buffer, -1, bBuffer, sizeof(bBuffer));

	LastOLEError = pCLSIDFromProgID(bBuffer, &CLSIDObj);
	if(SUCCEEDED(LastOLEError))
	{
		LastOLEError = pCoCreateInstance(CLSIDObj, NULL, CLSCTX_LOCAL_SERVER, IID_IDispatch, (void**)&pDisp);
		if(FAILED(LastOLEError))
			LastOLEError = pCoCreateInstance(CLSIDObj, NULL, CLSCTX_ALL, IID_IDispatch, (void**)&pDisp);
		if(SUCCEEDED(LastOLEError))
		{
			SETSV(1, CreatePerlObject(pPerl, pDisp, TRUE));
			bSuccess = TRUE;
		}
	}

	RETURNRESULT;
}

XS(Win32OLEDestroyObject)
{
	dXSARGS;
	LPWINOLEOBJECT lpObj;
	BOOL bSuccess = FALSE;

	if(items != 1)
	{
		CROAK("usage: Win32::OLEDestroyObject($object);\n");
	}
	OleStartupInit(pPerl);

	lpObj = GetOLEObject(pPerl, ST(0));
	if(IsOleStruct(lpObj))
	{
		ReleaseObjects(pPerl, lpObj);
			
		// unlink from list
		if(lpObj->pPrevious == NULL)
			g_lpObj = lpObj->pNext;
		else if(lpObj->pNext == NULL)
			lpObj->pPrevious->pNext = NULL;
		else
		{
			lpObj->pPrevious->pNext = lpObj->pNext;
			lpObj->pNext->pPrevious = lpObj->pPrevious;
		}

		Safefree(lpObj);
		SETIV(0,0);
		bSuccess = TRUE;
	}

	RETURNRESULT;
}

BOOL GetHashedDispID(CPerl *pPerl, LPWINOLEOBJECT lpObj, char *buffer, unsigned int length, DISPID &dispID)
{
	if(length == 0 || *buffer == '\0')
	{
		dispID = DISPID_VALUE;
		return TRUE;
	}

	if(!hv_exists(lpObj->hashTable, buffer, length))
	{	// not there so get if info and add it
		DISPID id;
		OLECHAR bBuffer[bufferSize], *pbBuffer;
		SV* sv;
		pbBuffer = bBuffer;
		MultiByteToWideChar(CP_ACP, NULL, buffer, -1, bBuffer, sizeof(bBuffer));
		LastOLEError = lpObj->pDispatch->GetIDsOfNames(IID_NULL, &pbBuffer, 1, lcidDefault, &id);
		if(SUCCEEDED(LastOLEError))
		{
			sv = newSViv(id);
			hv_store(lpObj->hashTable, buffer, length, sv, 0);
			dispID = id;
			return TRUE;
		}
	}
	else
	{
		SV** ppsv;
		ppsv = hv_fetch(lpObj->hashTable, buffer, length, 0);
		if(ppsv != NULL)
		{
			dispID = (DISPID)SvIV(*ppsv);
			return TRUE;
		}
		LastOLEError = INTERNAL_ERROR;
	}
	return FALSE;
}

void CreateSafeBinaryArray(CPerl *pPerl, SV* sv, VARIANT *pVariant)
{
	unsigned char* ptr;
	unsigned int length;
	SAFEARRAYBOUND rgsabound;
	unsigned char* pDest;

	ptr = (unsigned char*)SvPV(sv, length);
	rgsabound.lLbound = 0;
	rgsabound.cElements = length;
	pVariant->parray = pSafeArrayCreate(VT_UI1, 1, &rgsabound);
	if(pVariant->parray != NULL)
	{
		pVariant->vt = VT_UI1 | VT_ARRAY;
		HRESULT hResult = pSafeArrayAccessData(pVariant->parray, (void**)&pDest);
		if(SUCCEEDED(hResult))
		{
			memcpy(pDest, ptr, length);
			pSafeArrayUnaccessData(pVariant->parray);
		}
	}
}

void CreateSafeArray(CPerl *pPerl, AV* av, VARIANT *pVariant)
{
	SV **psv;
	char *ptr;
	unsigned int length;
	SAFEARRAYBOUND rgsabound;
	long arrayIndex;
	VARIANT variant;

	rgsabound.lLbound = 0;
	rgsabound.cElements = av_len(av)+1;
	pVariant->parray = pSafeArrayCreate(VT_VARIANT, 1, &rgsabound);
	if(pVariant->parray != NULL)
	{
		pVariant->vt = VT_VARIANT | VT_ARRAY;
		pVariantInit(&variant);
		variant.vt = VT_BSTR;
		for(arrayIndex = 0; arrayIndex < rgsabound.cElements; ++arrayIndex)
		{
			psv = av_fetch(av, arrayIndex, 0);
			if(psv != NULL)
			{
				ptr = SvPV(*psv, length);
				variant.bstrVal = AllocOLEString(ptr, length);
				if(variant.bstrVal != NULL)
					pSafeArrayPutElement(pVariant->parray, &arrayIndex, &variant);
			}
		}
	}
}

void DestroySafeArray(VARIANT *pVariant)
{
	long arrayIndex, upperBound;
	VARIANT variant;
	HRESULT hResult;

	hResult = pSafeArrayGetUBound(pVariant->parray, 1, &upperBound);
	if(SUCCEEDED(hResult))
	{
		for(arrayIndex = 0; arrayIndex <= upperBound; ++arrayIndex)
		{
			hResult = pSafeArrayGetElement(pVariant->parray, &arrayIndex, &variant);
			if(SUCCEEDED(hResult))
			{
				if(variant.vt == VT_BSTR)
				{
					pSysFreeString(variant.bstrVal);
					variant.bstrVal = NULL;
					pSafeArrayPutElement(pVariant->parray, &arrayIndex, &variant);
				}
			}
		}
	}
	pSafeArrayDestroy(pVariant->parray);
	pVariant->parray = NULL;
}

static char *IVType = "Type";
static const int IVTypeLength = strlen(IVType);
static char *IVValue = "Value";
static const int IVValueLength = strlen(IVValue);

void CreateVariantFromInternalVariant(CPerl *pPerl, SV* sv, VARIANT *pVariant)
{
	char *ptr;
	unsigned int length;
	int nType;
	SV **psv;
	HV* hv = (HV*)SvRV(sv);

	psv = hv_fetch(hv, IVType, IVTypeLength, 0);
	if(psv != NULL)
	{
		nType = SvIV(*psv);
		psv = hv_fetch(hv, IVValue, IVValueLength, 0);
		if(psv != NULL)
		{
			switch(nType)
			{
				case VT_UI1:
					switch(SvTYPE(*psv))
					{
						case SVt_PVIV:
						case SVt_PV:
							CreateSafeBinaryArray(pPerl, *psv, pVariant);
							break;

						default:
							V_VT(pVariant) = VT_UI1;
							V_UI1(pVariant) = SvIV(*psv);
							break;
					}
					break;

				case VT_BOOL:
					V_VT(pVariant) = VT_BOOL;
					V_BOOL(pVariant) = SvIV(*psv);
					break;

				case VT_I2:
					V_VT(pVariant) = VT_I2;
					V_I2(pVariant) = SvIV(*psv);
					break;

				case VT_I4:
					V_VT(pVariant) = VT_I4;
					V_I4(pVariant) = SvIV(*psv);
					break;

				case VT_R4:
					V_VT(pVariant) = VT_R4;
					V_R4(pVariant) = SvNV(*psv);
					break;

				case VT_R8:
					V_VT(pVariant) = VT_R8;
					V_R8(pVariant) = SvNV(*psv);
					break;

				case VT_BSTR:
					V_VT(pVariant) = VT_BSTR;
					ptr = SvPV(*psv, length);
					V_BSTR(pVariant) = AllocOLEString(ptr, length);
					break;

				case VT_DATE:
				case VT_CY:
					V_VT(pVariant) = VT_BSTR;
					ptr = SvPV(*psv, length);
					V_BSTR(pVariant) = AllocOLEString(ptr, length);
					pVariantChangeType(pVariant, pVariant, 0, nType);
					break;
			}
		}
	}
}

void CreateVariantFromSV(CPerl *pPerl, SV* sv, VARIANT *pVariant)
{
	char *ptr;
	unsigned int length;
	int type;

	pVariantInit(pVariant);

	if(SvROK(sv))
	{
		if(sv_isa(sv, "OLE"))
		{
			LPWINOLEOBJECT lpObj;
			lpObj = GetOLEObject(pPerl, sv);
			if(ValidDispatch(lpObj) && lpObj->hashTable != NULL)
			{
				V_VT(pVariant) = VT_DISPATCH;
				V_DISPATCH(pVariant) = lpObj->pDispatch;
				return;
			}
		}
		else if(sv_isa(sv, "OLE::Variant"))
		{
			CreateVariantFromInternalVariant(pPerl, sv, pVariant);
			return;
		}
		sv = SvRV(sv);
	}

	type = SvTYPE(sv);
	if(type == SVt_PVMG)
	{	// blessed scalar
		if(SvPOKp(sv))
			type = SVt_PV;
		else if(SvNOKp(sv))
			type = SVt_NV;
		else if(SvIOKp(sv))
			type = SVt_IV;
	}

	switch(type)
	{
		case SVt_PVAV:
			CreateSafeArray(pPerl, (AV*)sv, pVariant);
			break;

		case SVt_PVIV:
		case SVt_PV:
			pVariant->vt = VT_BSTR; 
			ptr = SvPV(sv, length);
			pVariant->bstrVal = AllocOLEString(ptr, length);
			break;

		case SVt_NV:
			pVariant->vt = VT_R8;
			pVariant->dblVal = SvNV(sv);
			break;

		default:
			pVariant->vt = VT_I4;
			pVariant->lVal = SvIV(sv);
			break;
	}
}

#define SETiVRETURN(x,f)\
					if(x->vt&VT_BYREF) {\
						sv_setiv(sv, (long)*x->p##f);\
					} else {\
						sv_setiv(sv, (long)x->f);\
					}

#define SETnVRETURN(x,f)\
					if(x->vt&VT_BYREF) {\
						sv_setnv(sv, (double)*x->p##f);\
					} else {\
						sv_setnv(sv, (double)x->f);\
					}

#undef bool

void SVFromVariant(CPerl *pPerl, VARIANT *pVariant, SV* sv)
{
	switch(pVariant->vt&~VT_BYREF)
	{
		case VT_EMPTY:
		case VT_NULL:
			sv_setsv(sv, &sv_undef);
			break;

		case VT_UI1:
			SETiVRETURN(pVariant,bVal)
			break;

		case VT_I2:
			SETiVRETURN(pVariant,iVal)
			break;

		case VT_I4:
			SETiVRETURN(pVariant,lVal)
			break;

		case VT_R4:
			SETnVRETURN(pVariant,fltVal)
			break;

		case VT_R8:
			SETnVRETURN(pVariant,dblVal)
			break;

		case VT_BSTR:
ConvertString:
			{
				int length;
				char *pStr;
				if(pVariant->vt&VT_BYREF)
					length = pSysStringLen(*pVariant->pbstrVal)+2;
				else
					length = pSysStringLen(pVariant->bstrVal)+2;

				NEW(1110, pStr, length, char);

				if(pVariant->vt&VT_BYREF)
					WideCharToMultiByte(CP_ACP, NULL, *pVariant->pbstrVal, -1, pStr, length, NULL, NULL);
				else
					WideCharToMultiByte(CP_ACP, NULL, pVariant->bstrVal, -1, pStr, length, NULL, NULL);

				sv_setpv(sv, pStr);
				// mega kludge -- check this in 5.003_11
				if(SvTYPE(sv) == SVt_PVNV)
					(sv)->sv_flags &= ~SVt_NV;

				Safefree(pStr);
			}
			break;

		case VT_ERROR:
			SETiVRETURN(pVariant,scode)
			break;

		case VT_BOOL:
			if(pVariant->vt&VT_BYREF)
			{
				sv_setiv(sv, (long)V_BOOLREF(pVariant));
			}
			else
			{
				sv_setiv(sv, (long)V_BOOL(pVariant));
			}
			break;

		case VT_DISPATCH:
			{
				IDispatch *pDisp;
				if(pVariant->vt&VT_BYREF) 
				{
					pDisp = (IDispatch*)*pVariant->ppunkVal;
				}
				else
				{
					pDisp = (IDispatch*)pVariant->punkVal;
				}
				sv_setsv(sv, CreatePerlObject(pPerl, pDisp));
			}
			break;

		case VT_UNKNOWN:
			{
				IUnknown *punk;
				IDispatch *pDisp;
				if(pVariant->vt&VT_BYREF) 
				{
					punk = (IUnknown*)*pVariant->ppunkVal;
				}
				else
				{
					punk = (IUnknown*)pVariant->punkVal;
				}
				if(SUCCEEDED(punk->QueryInterface(IID_IDispatch, (void**)&pDisp)))
					sv_setsv(sv, CreatePerlObject(pPerl, pDisp));
				else
					sv_setsv(sv, &sv_undef);
				punk->Release();
			}
			break;

		case VT_DATE:
		case VT_CY:
		case VT_VARIANT:
		default:
			{
				HRESULT hResult = pVariantChangeType(pVariant, pVariant, 0, VT_BSTR);
				if(SUCCEEDED(hResult))
					goto ConvertString;
			}

			sv_setsv(sv, &sv_undef);
			break;
	}
}

void SetSVFromVariant(CPerl *pPerl, VARIANT *pVariant, SV* sv)
{
	if(pVariant->vt&VT_ARRAY)
	{	// array of items
		SV *nsv;
		AV *av;
		VARIANT variant;
		int dim, index;
		long *pArrayIndex, *pLowerBound, *pUpperBound;
		HRESULT hResult;

		dim = pSafeArrayGetDim(pVariant->parray);
		NEW(4444, pArrayIndex, dim, long);
		NEW(4444, pLowerBound, dim, long);
		NEW(4444, pUpperBound, dim, long);
		for(index = 1; index <= dim; ++index)
		{
			hResult = pSafeArrayGetLBound(pVariant->parray, index, &pLowerBound[index-1]);
			if(FAILED(hResult))
				goto ErrorExit;
		}

		for(index = 1; index <= dim; ++index)
		{
			hResult = pSafeArrayGetUBound(pVariant->parray, index, &pUpperBound[index-1]);
			if(FAILED(hResult))
				goto ErrorExit;
		}

		av = newAV();
		if(dim < 3)
		{
			memcpy(pArrayIndex, pLowerBound, dim*sizeof(long));
			for(index = dim-1; pArrayIndex[index] <= pUpperBound[index]; ++pArrayIndex[index])
			{
				hResult = pSafeArrayGetElement(pVariant->parray, pArrayIndex, &variant);
				if(SUCCEEDED(hResult))
				{
					nsv = newSVsv(&sv_undef);
					SetSVFromVariant(pPerl, &variant, nsv);
					av_push(av, nsv);
				}
			}
		}
		sv_setsv(sv, newRV((SV*)av));

ErrorExit:
		Safefree(pArrayIndex);
		Safefree(pLowerBound);
		Safefree(pUpperBound);
	}
	else
		SVFromVariant(pPerl, pVariant, sv);
}

const int returnIndex = 2;

XS(Win32OLEDispatch)
{
	dXSARGS;
	char *buffer;
	char *ptr;
	unsigned int length, argErr;
	int index, arrayIndex;
	LPWINOLEOBJECT lpObj;
	EXCEPINFO excepinfo;
	DISPID dispID;
	VARIANT result;
	DISPPARAMS dispParams;
	BOOL bSuccess = FALSE;

	if(items < 3)
	{
		CROAK("usage: Win32::OLEDispatch($object, $funcName, $funcReturn, ...);\n");
	}
	OleStartupInit(pPerl);

	lpObj = GetOLEObject(pPerl, ST(0));
	if(ValidDispatch(lpObj) && lpObj->hashTable != NULL)
	{
		pVariantInit(&result);

		dispParams.rgvarg = NULL;
		dispParams.rgdispidNamedArgs = NULL;
		dispParams.cNamedArgs = 0;
		dispParams.cArgs = items - 3;

		if(dispParams.cArgs > 0)
		{
			NEW(2101, dispParams.rgvarg, dispParams.cArgs, VARIANTARG);
			for(index = 0; index < dispParams.cArgs; ++index)
				CreateVariantFromSV(pPerl, ST(items - 1 - index), &dispParams.rgvarg[index]);
		}

		buffer = SvPV(ST(1), length);
		if(GetHashedDispID(pPerl, lpObj, buffer, length, dispID))
		{
			EXCEPTIONINFOCLEAR(excepinfo);

			LastOLEError = lpObj->pDispatch->Invoke(dispID, IID_NULL, lcidDefault,
								DISPATCH_METHOD | DISPATCH_PROPERTYGET, &dispParams, &result, &excepinfo, &argErr);

			if(FAILED(LastOLEError))
			{	// mega kludge
				// if a method in WORD is called and we ask for a result when one is not returned
				// then hResult == DISP_E_EXCEPTION. this only happens on functions whose DISPID > 0x8000
				EXCEPTIONINFO(LastOLEError, excepinfo);

				if(LastOLEError == DISP_E_EXCEPTION && dispID > 0x8000)
				{
					EXCEPTIONINFOCLEAR(excepinfo);

					pVariantClear(&result);
					LastOLEError = lpObj->pDispatch->Invoke(dispID, IID_NULL, lcidDefault,
								DISPATCH_METHOD | DISPATCH_PROPERTYGET, &dispParams, NULL, &excepinfo, &argErr);

					EXCEPTIONINFO(LastOLEError, excepinfo);
					if(SUCCEEDED(LastOLEError))
						bSuccess = TRUE;
				}
			}
			else
				bSuccess = TRUE;
		}


		if(bSuccess)
		{	// handle result
			SetSVFromVariant(pPerl, &result, ST(returnIndex));
		}

		pVariantClear(&result);
		if(dispParams.cArgs != 0)
		{
			for(index = 0; index < dispParams.cArgs; ++index)
			{
				if(dispParams.rgvarg[index].vt == VT_BSTR)
					pSysFreeString(dispParams.rgvarg[index].bstrVal);

				else if(dispParams.rgvarg[index].vt == (VT_VARIANT | VT_ARRAY))
					DestroySafeArray(&dispParams.rgvarg[index]);
			}
			Safefree(dispParams.rgvarg);
		}
	}

	RETURNRESULT;
}

XS(Win32OLEGetProperty)
{
	dXSARGS;
	char *buffer;
	unsigned int length, argErr;
	int index;
	LPWINOLEOBJECT lpObj;
	EXCEPINFO excepinfo;
	DISPPARAMS dispParams;
	VARIANT result;
	DISPID dispID;
	BOOL bSuccess = FALSE;

	if(items < 3)
	{
		CROAK("usage: Win32::OLEGetProperty($object, $varName, $varReturn, ...);\n");
	}
	OleStartupInit(pPerl);

	lpObj = GetOLEObject(pPerl, ST(0));
	if(ValidDispatch(lpObj))
	{
		pVariantInit(&result);

		dispParams.rgvarg = NULL;
		dispParams.rgdispidNamedArgs = NULL;
		dispParams.cNamedArgs = 0;
		dispParams.cArgs = items - 3;
		if(dispParams.cArgs > 0)
		{
			NEW(2101, dispParams.rgvarg, dispParams.cArgs, VARIANTARG);
			for(index = 0; index < dispParams.cArgs; ++index)
			{
				pVariantInit(&dispParams.rgvarg[index]);
				dispParams.rgvarg[index].vt = VT_BSTR; 
				buffer = SvPV(ST(items - 1 - index), length);
				dispParams.rgvarg[index].bstrVal = AllocOLEString(buffer, length);
			}
				
		}

		buffer = SvPV(ST(1), length);
		if(GetHashedDispID(pPerl, lpObj, buffer, length, dispID))
		{
			EXCEPTIONINFOCLEAR(excepinfo);

			LastOLEError = lpObj->pDispatch->Invoke(dispID, IID_NULL, lcidDefault, DISPATCH_METHOD | DISPATCH_PROPERTYGET,
												&dispParams, &result, &excepinfo, &argErr);

			EXCEPTIONINFO(LastOLEError, excepinfo);
			bSuccess = SUCCEEDED(LastOLEError);
		}

		if(bSuccess)
		{	// handle result
			SetSVFromVariant(pPerl, &result, ST(returnIndex));
			pVariantClear(&result);
		}
		if(dispParams.cArgs != 0)
		{
			for(index = 0; index < dispParams.cArgs; ++index)
			{
				if(dispParams.rgvarg[index].vt == VT_BSTR)
					pSysFreeString(dispParams.rgvarg[index].bstrVal);
			}
			Safefree(dispParams.rgvarg);
		}
	}

	RETURNRESULT;
}

XS(Win32OLESetProperty)
{
	dXSARGS;
	unsigned int length, argErr;
	char *buffer;
	int index;
	LPWINOLEOBJECT lpObj;
	EXCEPINFO excepinfo;
	DISPID dispID;
	DISPID dispidNamed;
	DISPPARAMS dispParams;
	BOOL bSuccess = FALSE;

	if(items < 3)
	{
		CROAK("usage: Win32::OLESetProperty($object, $varName, $varValue, ...);\n");
	}
	OleStartupInit(pPerl);

	lpObj = GetOLEObject(pPerl, ST(0));
	if(ValidDispatch(lpObj))
	{
		dispParams.rgvarg = NULL;
		dispParams.rgdispidNamedArgs = NULL;
		dispParams.cNamedArgs = 0;
		dispParams.cArgs = items - 2;

		if(dispParams.cArgs > 0)
		{
			NEW(2101, dispParams.rgvarg, dispParams.cArgs, VARIANTARG);
			dispParams.rgvarg[0].vt = VT_BSTR; 
			buffer = SvPV(ST(2), length);
			dispParams.rgvarg[0].bstrVal = AllocOLEString(buffer, length);

			if(dispParams.cArgs > 1)
			{
				dispidNamed = DISPID_PROPERTYPUT;
				dispParams.rgdispidNamedArgs = &dispidNamed;
				dispParams.cNamedArgs = 1;

				for(index = 1; index < dispParams.cArgs; ++index)
					CreateVariantFromSV(pPerl, ST(items - index), &dispParams.rgvarg[index]);
			}
		}


		buffer = SvPV(ST(1), length);
		if(GetHashedDispID(pPerl, lpObj, buffer, length, dispID))
		{
			EXCEPTIONINFOCLEAR(excepinfo);

			LastOLEError = lpObj->pDispatch->Invoke(dispID, IID_NULL, lcidDefault, DISPATCH_PROPERTYPUT,
												&dispParams, NULL, &excepinfo, &argErr);

			EXCEPTIONINFO(LastOLEError, excepinfo);
			bSuccess = SUCCEEDED(LastOLEError);
		}


		if(dispParams.cArgs > 0)
		{
			if(dispParams.rgvarg[0].vt == VT_BSTR)
				pSysFreeString(dispParams.rgvarg[0].bstrVal);

			Safefree(dispParams.rgvarg);
		}
	}

	RETURNRESULT;
}


SV *Win32OLEPropertyGet(CPerl *pPerl, SV *svOle_object, SV *PropertyName)
{
	char *buffer;
	unsigned int length, argErr;
	LPWINOLEOBJECT lpObj;
	EXCEPINFO excepinfo;
	DISPPARAMS dispParams;
	VARIANT result;
	DISPID dispID;
	BOOL bSuccess = FALSE;
	SV* sv = NULL;

	OleStartupInit(pPerl);
	lpObj = (LPWINOLEOBJECT)SvIV(svOle_object);
	if(ValidDispatch(lpObj))
	{
		pVariantInit(&result);

		dispParams.rgvarg = NULL;
		dispParams.rgdispidNamedArgs = NULL;
		dispParams.cNamedArgs = 0;
		dispParams.cArgs = 0;

		buffer = SvPV(PropertyName, length);
		if(GetHashedDispID(pPerl, lpObj, buffer, length, dispID))
		{
			EXCEPTIONINFOCLEAR(excepinfo);

			LastOLEError = lpObj->pDispatch->Invoke(dispID, IID_NULL, lcidDefault, DISPATCH_METHOD | DISPATCH_PROPERTYGET,
												&dispParams, &result, &excepinfo, &argErr);

			EXCEPTIONINFO(LastOLEError, excepinfo);
			bSuccess = SUCCEEDED(LastOLEError);
		}

		if(bSuccess)
		{	// handle result
			sv = newSVsv(&sv_undef);
			SetSVFromVariant(pPerl, &result, sv);
			pVariantClear(&result);
		}
	}
	return sv;
}

SV *Win32OLEPropertySet(CPerl *pPerl, SV *svOle_object, SV *PropertyName, SV *Value)
{
	unsigned int length, argErr;
	char *buffer;
	int index;
	LPWINOLEOBJECT lpObj;
	EXCEPINFO excepinfo;
	DISPID dispID, dispIDParam;
	DISPPARAMS dispParams;
	VARIANT propertyValue;
	BOOL bSuccess = FALSE;

	OleStartupInit(pPerl);
	lpObj = (LPWINOLEOBJECT)SvIV(svOle_object);
	if(ValidDispatch(lpObj))
	{
		dispIDParam = DISPID_PROPERTYPUT;
		dispParams.rgvarg = &propertyValue;
		dispParams.rgdispidNamedArgs = &dispIDParam;
		dispParams.cNamedArgs = 1;
		dispParams.cArgs = 1;

		CreateVariantFromSV(pPerl, Value, &propertyValue);

		buffer = SvPV(PropertyName, length);
		if(GetHashedDispID(pPerl, lpObj, buffer, length, dispID))
		{
			EXCEPTIONINFOCLEAR(excepinfo);

			LastOLEError = lpObj->pDispatch->Invoke(dispID, IID_NULL, lcidDefault, DISPATCH_PROPERTYPUT,
												&dispParams, NULL, &excepinfo, &argErr);

			EXCEPTIONINFO(LastOLEError, excepinfo);
			bSuccess = SUCCEEDED(LastOLEError);
		}


		if(propertyValue.vt == VT_BSTR)
			pSysFreeString(propertyValue.bstrVal);
		else if(propertyValue.vt == (VT_VARIANT | VT_ARRAY))
			DestroySafeArray(&propertyValue);
	}

	return (bSuccess ? Value : NULL);
}


XS(XS_OLEhash_FETCH)
{
	dXSARGS;
	SV **coo;
	SV *key, *retval, *temp;

	if(items != 2)
	{
		CROAK("Usage: OLEhash::FETCH(self,key)");
	}

	retval = Nullsv;
	coo = hv_fetch((HV*)SvRV(ST(0)), OleObjectID, OleObjectIDLength, 0);
	key = ST(1);
	if(coo != NULL)
	{
		if(strcmp(SvPV(key, na), OleObjectID) == 0)
		{
			ST(0) = *coo;
		}
		else
		{
			temp = Win32OLEPropertyGet(pPerl, *coo, key);
			if(temp != NULL)
				retval = temp;

			ST(0) = retval;
			sv_2mortal(ST(0));
		}
	}

	XSRETURN(1);
}

XS(XS_OLEhash_STORE)
{
	dXSARGS;
	SV **coo;

	if(items != 3)
	{
		CROAK("Usage: OLEhash::STORE(self,key,value)");
	}

	coo = hv_fetch((HV*)SvRV(ST(0)), OleObjectID, OleObjectIDLength, 0);
	if(coo != NULL)
		Win32OLEPropertySet(pPerl, *coo, ST(1), ST(2));

	ST(0) = newSVsv(&sv_undef);
	sv_2mortal(ST(0));
	XSRETURN(1);
}


XS(XS_OLEhash_FIRSTKEY)
{
	dXSARGS;
	SV **coo;
	unsigned int argErr;
	LPWINOLEOBJECT lpObj;
	EXCEPINFO excepinfo;
	DISPPARAMS dispParams;
	VARIANT result;
	BOOL bSuccess = FALSE;

	if(items != 1)
	{
		CROAK("Usage: OLEhash::FIRSTKEY(self)");
	}

	coo = hv_fetch((HV*)SvRV(ST(0)), OleObjectID, OleObjectIDLength, 0);
	if(coo != NULL)
	{
		ST(0) = newSVsv(&sv_undef);
		lpObj = (LPWINOLEOBJECT)SvIV(*coo);
		if(ValidDispatch(lpObj))
		{
			pVariantInit(&result);

			dispParams.rgvarg = NULL;
			dispParams.rgdispidNamedArgs = NULL;
			dispParams.cNamedArgs = 0;
			dispParams.cArgs = 0;

			EXCEPTIONINFOCLEAR(excepinfo);

			LastOLEError = lpObj->pDispatch->Invoke(DISPID_NEWENUM, IID_NULL, lcidDefault, DISPATCH_METHOD | DISPATCH_PROPERTYGET,
												&dispParams, &result, &excepinfo, &argErr);

			EXCEPTIONINFO(LastOLEError, excepinfo);
			bSuccess = SUCCEEDED(LastOLEError);
			if(bSuccess)
			{
				if((result.vt&~VT_BYREF) == VT_UNKNOWN)
				{
					IUnknown *punk;
					IDispatch *pDisp;
					IEnumVARIANT *pEnum;
					if(result.vt&VT_BYREF) 
					{
						punk = (IUnknown*)*result.ppunkVal;
					}
					else
					{
						punk = (IUnknown*)result.punkVal;
					}
					if(SUCCEEDED(punk->QueryInterface(IID_IDispatch, (void**)&pDisp)))
						sv_setsv(ST(0), CreatePerlObject(pPerl, pDisp));
					else if(SUCCEEDED(punk->QueryInterface(IID_IEnumVARIANT, (void**)&pEnum)))
					{
						if(lpObj->pEnum != NULL)
							lpObj->pEnum->Release();

						lpObj->pEnum = pEnum;
						pVariantClear(&result);

						if(SUCCEEDED(pEnum->Reset()) && SUCCEEDED(pEnum->Next(1, &result, NULL)))
							SetSVFromVariant(pPerl, &result, ST(0));
						else
							sv_setsv(ST(0), &sv_undef);
					}
					else
						sv_setsv(ST(0), &sv_undef);

					punk->Release();
				}
				else
					SetSVFromVariant(pPerl, &result, ST(0));

				pVariantClear(&result);
			}
		}
		sv_2mortal(ST(0));
	}

	XSRETURN(1);
}


XS(XS_OLEhash_NEXTKEY)
{
	dXSARGS;
	SV **coo;
	LPWINOLEOBJECT lpObj;
	VARIANT result;
	BOOL bSuccess = FALSE;

	if(items != 2)
	{
		CROAK("Usage: OLEhash::NEXTKEY(self, lastKey)");
	}

	coo = hv_fetch((HV*)SvRV(ST(0)), OleObjectID, OleObjectIDLength, 0);
	if(coo != NULL)
	{
		ST(0) = newSVsv(&sv_undef);
		lpObj = (LPWINOLEOBJECT)SvIV(*coo);
		if(ValidDispatch(lpObj) && lpObj->pEnum != NULL)
		{
			pVariantInit(&result);
			bSuccess = lpObj->pEnum->Next(1, &result, NULL) == S_OK;
			// if the return is anything other than S_OK
			// then this interface is no longer needed
			if(!bSuccess)
			{
				lpObj->pEnum->Release();
				lpObj->pEnum = NULL;
			}
		}
	}

	if(bSuccess)
		SetSVFromVariant(pPerl, &result, ST(0));
	else
	{
		ST(0) = newSVsv(&sv_undef);
		sv_2mortal(ST(0));
	}

	XSRETURN(1);
}


void Win32OLEInit(CPerl *pPerl)
{
	newXSInternal("Win32::OLELastError",		Win32OLELastError,		thisFile);
	newXSInternal("Win32::OLECreateObject",		Win32OLECreateObject,	thisFile);
	newXSInternal("Win32::OLEDestroyObject",	Win32OLEDestroyObject,	thisFile);
	newXSInternal("Win32::OLEDispatch",			Win32OLEDispatch,		thisFile);
	newXSInternal("Win32::OLEGetProperty",		Win32OLEGetProperty,	thisFile);
	newXSInternal("Win32::OLESetProperty",		Win32OLESetProperty,	thisFile);

	newXSInternal("OLEhash::FETCH",		XS_OLEhash_FETCH, thisFile);
	newXSInternal("OLEhash::STORE",		XS_OLEhash_STORE, thisFile);
	newXSInternal("OLEhash::FIRSTKEY",	XS_OLEhash_FIRSTKEY, thisFile);
	newXSInternal("OLEhash::NEXTKEY",	XS_OLEhash_NEXTKEY, thisFile);
}

