// NTSock.C

// (c) 1995 Microsoft Corporation. All rights reserved. 
//	Developed by ActiveWare Internet Corp., http://www.ActiveWare.com
// Portions (c) 1993 Intergraph Corporation. All rights reserved.

//    You may distribute under the terms of either the GNU General Public
//    License or the Artistic License, as specified in the README file.

#include <windows.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <assert.h>
#include <sys/socket.h>
#define WIN32_LEAN_AND_MEAN
#include "EXTERN.h"
#include "perl.h"

#define CROAK pPerl->croak

#ifdef USE_SOCKETS_AS_HANDLES
// thanks to Beverly Brown	(beverly@datacube.com)
#define OPEN_SOCKET(x)	pVtbl->OpenOSfhandle(x,_O_RDWR|_O_BINARY)
#define TO_SOCKET(x)	pVtbl->GetOSfhandle(x)

#else

#define OPEN_SOCKET(x)	(x)
#define TO_SOCKET(x)	(x)

#endif	// USE_SOCKETS_AS_HANDLES

//
// This is a clone of fdopen so that we can handle the version of sockets that NT gets to use.
//
// The problem is that sockets are not real file handles and 
// cannot be fdopen'ed. This causes problems in the do_socket
// routine in doio.c, since it tries to create two file pointers
// for the socket just created. We'll fake out an fdopen and see
// if we can prevent perl from trying to do stdio on sockets.
//

#define SOCKETAPI PASCAL 

typedef SOCKET (SOCKETAPI *LPSOCKACCEPT)(SOCKET, struct sockaddr *, int *);
typedef int (SOCKETAPI *LPSOCKBIND)(SOCKET, const struct sockaddr *, int);
typedef int (SOCKETAPI *LPSOCKCLOSESOCKET)(SOCKET);
typedef int (SOCKETAPI *LPSOCKCONNECT)(SOCKET, const struct sockaddr *, int);
typedef int (SOCKETAPI *LPSOCKIOCTLSOCKET)(SOCKET, long, u_long *);
typedef int (SOCKETAPI *LPSOCKGETPEERNAME)(SOCKET, struct sockaddr *, int *);
typedef int (SOCKETAPI *LPSOCKGETSOCKNAME)(SOCKET, struct sockaddr *, int *);
typedef int (SOCKETAPI *LPSOCKGETSOCKOPT)(SOCKET, int, int, char *, int *);
typedef u_long (SOCKETAPI *LPSOCKHTONL)(u_long);
typedef u_short (SOCKETAPI *LPSOCKHTONS)(u_short);
typedef int (SOCKETAPI *LPSOCKLISTEN)(SOCKET, int);
typedef u_long (SOCKETAPI *LPSOCKNTOHL)(u_long);
typedef u_short (SOCKETAPI *LPSOCKNTOHS)(u_short);
typedef int (SOCKETAPI *LPSOCKRECV)(SOCKET, char *, int, int);
typedef int (SOCKETAPI *LPSOCKRECVFROM)(SOCKET, char *, int, int, struct sockaddr *, int *);
typedef int (SOCKETAPI *LPSOCKSELECT)(int, fd_set *, fd_set *, fd_set *, const struct timeval *);
typedef int (SOCKETAPI *LPSOCKSEND)(SOCKET, const char *, int, int);
typedef int (SOCKETAPI *LPSOCKSENDTO)(SOCKET, const char *, int, int, const struct sockaddr *, int);
typedef int (SOCKETAPI *LPSOCKSETSOCKOPT)(SOCKET, int, int, const char *, int);
typedef int (SOCKETAPI *LPSOCKSHUTDOWN)(SOCKET, int);
typedef SOCKET (SOCKETAPI *LPSOCKSOCKET)(int, int, int);

/* Database function prototypes */
typedef struct hostent *(SOCKETAPI *LPSOCKGETHOSTBYADDR)(const char *, int, int);
typedef struct hostent *(SOCKETAPI *LPSOCKGETHOSTBYNAME)(const char *);
typedef int (SOCKETAPI *LPSOCKGETHOSTNAME)(char *, int);
typedef struct servent *(SOCKETAPI *LPSOCKGETSERVBYPORT)(int, const char *);
typedef struct servent *(SOCKETAPI *LPSOCKGETSERVBYNAME)(const char *, const char *);
typedef struct protoent *(SOCKETAPI *LPSOCKGETPROTOBYNUMBER)(int);
typedef struct protoent *(SOCKETAPI *LPSOCKGETPROTOBYNAME)(const char *);

/* Microsoft Windows Extension function prototypes */
typedef int (SOCKETAPI *LPSOCKWSASTARTUP)(unsigned short, LPWSADATA);
typedef int (SOCKETAPI *LPSOCKWSACLEANUP)(void);
typedef int (SOCKETAPI *LPSOCKWSAGETLASTERROR)(void);
typedef int (SOCKETAPI *LPWSAFDIsSet)(SOCKET, fd_set *);

static HINSTANCE hWinSockDll = 0;
static BOOL bInitialized = FALSE;
extern CRITICAL_SECTION csSock;

static LPSOCKACCEPT paccept = 0;
static LPSOCKBIND pbind = 0;
static LPSOCKCLOSESOCKET pclosesocket = 0;
static LPSOCKCONNECT pconnect = 0;
static LPSOCKIOCTLSOCKET pioctlsocket = 0;
static LPSOCKGETPEERNAME pgetpeername = 0;
static LPSOCKGETSOCKNAME pgetsockname = 0;
static LPSOCKGETSOCKOPT pgetsockopt = 0;
static LPSOCKHTONL phtonl = 0;
static LPSOCKHTONS phtons = 0;
static LPSOCKLISTEN plisten = 0;
static LPSOCKNTOHL pntohl = 0;
static LPSOCKNTOHS pntohs = 0;
static LPSOCKRECV precv = 0;
static LPSOCKRECVFROM precvfrom = 0;
static LPSOCKSELECT pselect = 0;
static LPSOCKSEND psend = 0;
static LPSOCKSENDTO psendto = 0;
static LPSOCKSETSOCKOPT psetsockopt = 0;
static LPSOCKSHUTDOWN pshutdown = 0;
static LPSOCKSOCKET psocket = 0;
static LPSOCKGETHOSTBYADDR pgethostbyaddr = 0;
static LPSOCKGETHOSTBYNAME pgethostbyname = 0;
static LPSOCKGETHOSTNAME pgethostname = 0;
static LPSOCKGETSERVBYPORT pgetservbyport = 0;
static LPSOCKGETSERVBYNAME pgetservbyname = 0;
static LPSOCKGETPROTOBYNUMBER pgetprotobynumber = 0;
static LPSOCKGETPROTOBYNAME pgetprotobyname = 0;
static LPSOCKWSASTARTUP pWSAStartup = 0;
static LPSOCKWSACLEANUP pWSACleanup = 0;
static LPSOCKWSAGETLASTERROR pWSAGetLastError = 0;
static LPWSAFDIsSet pWSAFDIsSet = 0;

void *GetAddress(CPerl *pPerl, HINSTANCE hInstance, char *lpFunctionName)
{
	char buffer[512];
	FARPROC proc = GetProcAddress(hInstance, lpFunctionName);
	if(proc == 0)
	{
		sprintf(buffer, "Unable to get address of %s in WSock32.dll", lpFunctionName);
		CROAK(buffer);
	}
	return proc;
}

void LoadWinSock(CPerl *pPerl)
{
	EnterCriticalSection(&csSock);
	if(hWinSockDll == NULL)
	{
		HINSTANCE hLib = LoadLibrary("WSock32.DLL");
		if(hLib == NULL)
			CROAK("Could not load WSock32.dll\n");

		paccept = (LPSOCKACCEPT)GetAddress(pPerl, hLib, "accept");
		pbind = (LPSOCKBIND)GetAddress(pPerl, hLib, "bind");
		pclosesocket = (LPSOCKCLOSESOCKET)GetAddress(pPerl, hLib, "closesocket");
		pconnect = (LPSOCKCONNECT)GetAddress(pPerl, hLib, "connect");
		pioctlsocket = (LPSOCKIOCTLSOCKET)GetAddress(pPerl, hLib, "ioctlsocket");
		pgetpeername = (LPSOCKGETPEERNAME)GetAddress(pPerl, hLib, "getpeername");
		pgetsockname = (LPSOCKGETSOCKNAME)GetAddress(pPerl, hLib, "getsockname");
		pgetsockopt = (LPSOCKGETSOCKOPT)GetAddress(pPerl, hLib, "getsockopt");
		phtonl = (LPSOCKHTONL)GetAddress(pPerl, hLib, "htonl");
		phtons = (LPSOCKHTONS)GetAddress(pPerl, hLib, "htons");
		plisten = (LPSOCKLISTEN)GetAddress(pPerl, hLib, "listen");
		pntohl = (LPSOCKNTOHL)GetAddress(pPerl, hLib, "ntohl");
		pntohs = (LPSOCKNTOHS)GetAddress(pPerl, hLib, "ntohs");
		precv = (LPSOCKRECV)GetAddress(pPerl, hLib, "recv");
		precvfrom = (LPSOCKRECVFROM)GetAddress(pPerl, hLib, "recvfrom");
		pselect = (LPSOCKSELECT)GetAddress(pPerl, hLib, "select");
		psend = (LPSOCKSEND)GetAddress(pPerl, hLib, "send");
		psendto = (LPSOCKSENDTO)GetAddress(pPerl, hLib, "sendto");
		psetsockopt = (LPSOCKSETSOCKOPT)GetAddress(pPerl, hLib, "setsockopt");
		pshutdown = (LPSOCKSHUTDOWN)GetAddress(pPerl, hLib, "shutdown");
		psocket = (LPSOCKSOCKET)GetAddress(pPerl, hLib, "socket");
		pgethostbyaddr = (LPSOCKGETHOSTBYADDR)GetAddress(pPerl, hLib, "gethostbyaddr");
		pgethostbyname = (LPSOCKGETHOSTBYNAME)GetAddress(pPerl, hLib, "gethostbyname");
		pgethostname = (LPSOCKGETHOSTNAME)GetAddress(pPerl, hLib, "gethostname");
		pgetservbyport = (LPSOCKGETSERVBYPORT)GetAddress(pPerl, hLib, "getservbyport");
		pgetservbyname = (LPSOCKGETSERVBYNAME)GetAddress(pPerl, hLib, "getservbyname");
		pgetprotobynumber = (LPSOCKGETPROTOBYNUMBER)GetAddress(pPerl, hLib, "getprotobynumber");
		pgetprotobyname = (LPSOCKGETPROTOBYNAME)GetAddress(pPerl, hLib, "getprotobyname");
		pWSAStartup = (LPSOCKWSASTARTUP)GetAddress(pPerl, hLib, "WSAStartup");
		pWSACleanup = (LPSOCKWSACLEANUP)GetAddress(pPerl, hLib, "WSACleanup");
		pWSAGetLastError = (LPSOCKWSAGETLASTERROR)GetAddress(pPerl, hLib, "WSAGetLastError");
		pWSAFDIsSet = (LPWSAFDIsSet)GetAddress(pPerl, hLib, "__WSAFDIsSet");
		hWinSockDll = hLib;
	}
	LeaveCriticalSection(&csSock);
}

void EndSockets(void)
{
	if(hWinSockDll != NULL)
	{
		if(bInitialized)
			pWSACleanup();
		FreeLibrary(hWinSockDll);
	}
	hWinSockDll = NULL;
}

void StartSockets(CPerl *pPerl) 
{
	unsigned short version;
	WSADATA retdata;
	int ret;
	int iSockOpt = SO_SYNCHRONOUS_NONALERT;

	LoadWinSock(pPerl);
	//
	// initalize the winsock interface and insure that it is
	// cleaned up at exit.
	//
	version = 0x101;
	if(ret = pWSAStartup(version, &retdata))
		CROAK("Unable to locate winsock library!\n");
	bInitialized = TRUE;
	if(retdata.wVersion != version)
		CROAK("Could not find version 1.1 of winsock dll\n");

	// atexit((void (*)(void)) EndSockets);

#ifdef USE_SOCKETS_AS_HANDLES
	//
	// Enable the use of sockets as filehandles
	//
	psetsockopt(INVALID_SOCKET, SOL_SOCKET, SO_OPENTYPE, (char *)&iSockOpt, sizeof(iSockOpt));
#endif	// USE_SOCKETS_AS_HANDLES
}


#ifndef USE_SOCKETS_AS_HANDLES
FILE *myfdopen(int fd, char *mode)
{
	FILE *fp;
	char sockbuf[256];
	int optlen = sizeof(sockbuf);
	int retval;

	
	if(hWinSockDll == 0)
		LoadWinSock();

	retval = pgetsockopt((SOCKET)fd, SOL_SOCKET, SO_TYPE, sockbuf, &optlen);
	if(retval == SOCKET_ERROR && pWSAGetLastError() == WSAENOTSOCK) 
	{
		return(_fdopen(fd, mode));
	}

	//
	// If we get here, then fd is actually a socket.
	//
	Newz(1601, fp, 1, FILE);
	if(fp == NULL)
	{
		errno = ENOMEM;
		return NULL;
	}

	fp->_file = fd;
	if(*mode == 'r')
		fp->_flag = _IOREAD;
	else
		fp->_flag = _IOWRT;

	return fp;
}
#endif	// USE_SOCKETS_AS_HANDLES


u_long nt_htonl(u_long hostlong)
{
	if(hWinSockDll == 0)
		LoadWinSock(this);

	return phtonl(hostlong);
}

u_short nt_htons(u_short hostshort)
{
	if(hWinSockDll == 0)
		LoadWinSock(this);

	return phtons(hostshort);
}

u_long nt_ntohl(u_long netlong)
{
	if(hWinSockDll == 0)
		LoadWinSock(this);

	return pntohl(netlong);
}

u_short nt_ntohs(u_short netshort)
{
	if(hWinSockDll == 0)
		LoadWinSock(this);

	return pntohs(netshort);
}


#define SOCKET_TEST(x, y)	if(bInitialized == 0) StartSockets(this);\
							if((x) == (y)) errno = pWSAGetLastError()

#define SOCKET_TEST_ERROR(x) SOCKET_TEST(x,	SOCKET_ERROR)

SOCKET nt_accept(SOCKET s, struct sockaddr *addr, int *addrlen)
{
	SOCKET r;

	SOCKET_TEST((r = paccept(TO_SOCKET(s), addr, addrlen)), INVALID_SOCKET);
	return OPEN_SOCKET(r);
}

int nt_bind(SOCKET s, struct sockaddr *addr, int addrlen)
{
	int r;

	SOCKET_TEST_ERROR(r = pbind(TO_SOCKET(s), addr, addrlen));
	return r;
}

int nt_connect(SOCKET s, struct sockaddr *addr, int addrlen)
{
	int r;

	SOCKET_TEST_ERROR(r = pconnect(TO_SOCKET(s), addr, addrlen));
	return r;
}


int nt_getpeername(SOCKET s, struct sockaddr *addr, int *addrlen)
{
	int r;

	SOCKET_TEST_ERROR(r = pgetpeername(TO_SOCKET(s), addr, addrlen));
	return r;
}

int nt_getsockname(SOCKET s, struct sockaddr *addr, int *addrlen)
{
	int r;

	SOCKET_TEST_ERROR(r = pgetsockname(TO_SOCKET(s), addr, addrlen));
	return r;
}

int nt_getsockopt(SOCKET s, int level, int optname, char *optval, int *optlen)
{
	int r;

	SOCKET_TEST_ERROR(r = pgetsockopt(TO_SOCKET(s), level, optname, optval, optlen));
	return r;
}

int nt_ioctlsocket(SOCKET s, long cmd, u_long *argp)
{
	int r;

	SOCKET_TEST_ERROR(r = pioctlsocket(TO_SOCKET(s), cmd, argp));
	return r;
}

int nt_listen(SOCKET s, int backlog)
{
	int r;

	SOCKET_TEST_ERROR(r = plisten(TO_SOCKET(s), backlog));
	return r;
}

int nt_recv(SOCKET s, char *buf, int len, int flags)
{
	int r;

	SOCKET_TEST_ERROR(r = precv(TO_SOCKET(s), buf, len, flags));
	return r;
}

int nt_recvfrom(SOCKET s, char *buf, int len, int flags, struct sockaddr *from, int *fromlen)
{
	int r;

	SOCKET_TEST_ERROR(r = precvfrom(TO_SOCKET(s), buf, len, flags, from, fromlen));
	return r;
}

// select contributed by Vincent R. Slyngstad (vrs@ibeam.intel.com)
long nt_select(int nfds, int* rd, int* wr, int* ex, struct timeval* timeout)
{
	long r;
	int i, fd, bit, offset;
	FD_SET nrd, nwr, nex;

	FD_ZERO(&nrd);
	FD_ZERO(&nwr);
	FD_ZERO(&nex);
	for (i = 0; i < nfds; i++) 
	{
		fd = TO_SOCKET(i);
		bit = 1L<<(i % (sizeof(int)*8));
		offset = i / (sizeof(int)*8);
		if(rd != NULL && (rd[offset] & bit))
			FD_SET(fd, &nrd);
		if(wr != NULL && (wr[offset] & bit))
			FD_SET(fd, &nwr);
		if(ex != NULL && (ex[offset] & bit))
			FD_SET(fd, &nex);
	}

	SOCKET_TEST_ERROR(r = pselect(nfds, &nrd, &nwr, &nex, timeout));

	for(i = 0; i < nfds; i++) 
	{
		fd = TO_SOCKET(i);
		bit = 1L<<(i % (sizeof(int)*8));
		offset = i / (sizeof(int)*8);
		if(rd != NULL && (rd[offset] & bit)) 
		{
			if(!pWSAFDIsSet(fd, &nrd))
				rd[offset] &= ~bit;
		}
		if(wr != NULL && (wr[offset] & bit)) 
		{
			if(!pWSAFDIsSet(fd, &nwr))
				wr[offset] &= ~bit;
		}
		if(ex != NULL && (ex[offset] & bit)) 
		{
			if(!pWSAFDIsSet(fd, &nex))
				ex[offset] &= ~bit;
		}
	}
	return r;
}

int nt_send(SOCKET s, char *buf, int len, int flags)
{
	int r;

	SOCKET_TEST_ERROR(r = psend(TO_SOCKET(s), buf, len, flags));
	return r;
}

int nt_sendto(SOCKET s, char *buf, int len, int flags, struct sockaddr *to, int tolen)
{
	int r;

	SOCKET_TEST_ERROR(r = psendto(TO_SOCKET(s), buf, len, flags, to, tolen));
	return r;
}

int nt_setsockopt(SOCKET s, int level, int optname, char *optval, int optlen)
{
	int r;

	SOCKET_TEST_ERROR(r = psetsockopt(TO_SOCKET(s), level, optname, optval, optlen));
	return r;
}
    
int nt_shutdown(SOCKET s, int how)
{
	int r;

	SOCKET_TEST_ERROR(r = pshutdown(TO_SOCKET(s), how));
	return r;
}

SOCKET nt_socket(int af, int type, int protocol)
{
	SOCKET s;

#ifndef USE_SOCKETS_AS_HANDLES
	SOCKET_TEST(s = psocket(af, type, protocol), INVALID_SOCKET);
#else
	if(bInitialized == 0)
		StartSockets(this);

	if((s = psocket(af, type, protocol)) == INVALID_SOCKET)
		errno = pWSAGetLastError();
	else
		s = OPEN_SOCKET(s);
#endif	// USE_SOCKETS_AS_HANDLES

	return s;
}

struct hostent *nt_gethostbyaddr(char *addr, int len, int type)
{
	struct hostent *r;

	SOCKET_TEST(r = pgethostbyaddr(addr, len, type), NULL);
	return r;
}

struct hostent *nt_gethostbyname(char *name)
{
	struct hostent *r;

	SOCKET_TEST(r = pgethostbyname(name), NULL);
	return r;
}

int nt_gethostname(char *name, int len)
{
	int r;

	SOCKET_TEST_ERROR(r = pgethostname(name, len));
	return r;
}

struct protoent *nt_getprotobyname(char *name)
{
	struct protoent *r;

	SOCKET_TEST(r = pgetprotobyname(name), NULL);
	return r;
}

struct protoent *nt_getprotobynumber(int num)
{
	struct protoent *r;

	SOCKET_TEST(r = pgetprotobynumber(num), NULL);
	return r;
}

struct servent *nt_getservbyname(char *name, char *proto)
{
	struct servent *r;

	SOCKET_TEST(r = pgetservbyname(name, proto), NULL);
	return r;
}

struct servent *nt_getservbyport(int port, char *proto)
{
	struct servent *r;

	SOCKET_TEST(r = pgetservbyport(port, proto), NULL);
	return r;
}

//
// Networking stubs
//
#undef CROAK 
#define CROAK croak

void endhostent() 
{
	CROAK("endhostent not implemented!\n");
}

void endnetent()
{
	CROAK("endnetent not implemented!\n");
}

void endprotoent()
{
	CROAK("endprotoent not implemented!\n");
}

void endservent()
{
	CROAK("endservent not implemented!\n");
}


struct netent *getnetent(void) 
{
	CROAK("getnetent not implemented!\n");
	return (struct netent *) NULL;
}

struct netent *getnetbyname(char *name) 
{
	CROAK("getnetbyname not implemented!\n");
	return (struct netent *)NULL;
}

struct netent *getnetbyaddr(long net, int type) 
{
	CROAK("getnetbyaddr not implemented!\n");
	return (struct netent *)NULL;
}

struct protoent *getprotoent(void) 
{
	CROAK("getprotoent not implemented!\n");
	return (struct protoent *) NULL;
}

struct servent *getservent(void) 
{
	CROAK("getservent not implemented!\n");
	return (struct servent *) NULL;
}

void sethostent(int stayopen)
{
	CROAK("sethostent not implemented!\n");
}


void setnetent(int stayopen)
{
	CROAK("setnetent not implemented!\n");
}


void setprotoent(int stayopen)
{
	CROAK("setprotoent not implemented!\n");
}


void setservent(int stayopen)
{
	CROAK("setservent not implemented!\n");
}

int NtIsSocket(int fd)
{
	char sockbuf[256];
	int optlen = sizeof(sockbuf);
	int retval = 0;

	SOCKET_TEST_ERROR( retval = pgetsockopt( TO_SOCKET( fd ),
											 SOL_SOCKET, 
											 SO_TYPE, 
											 sockbuf, 
											 &optlen ));
					   
	if( retval == SOCKET_ERROR && errno == WSAENOTSOCK) 
	{
		retval = -1;
	}

	return retval;
}
