// PerlExe.cpp

// (c) 1995 Microsoft Corporation. All rights reserved. 
//	Developed by ActiveWare Internet Corp., http://www.ActiveWare.com

//    You may distribute under the terms of either the GNU General Public
//    License or the Artistic License, as specified in the README file.

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <io.h>
#include <errno.h>
#include <process.h>
#include <sys/stat.h>
#include <sys/utime.h>
#include <signal.h>
#include <direct.h>
#define WIN32_LEAN_AND_MEAN
#include "EXTERN.h"
#include "perl.h"
#include "perllib.h"
#include "IPerlSup.h"

#undef exit
#undef _exit
#undef errno
#undef malloc
#undef realloc
#undef free
#undef stdin
#undef stdout
#undef stderr
#define stderr (&_iob[2])
#undef clearerr
#undef fclose
#undef fdopen
#undef _fdopen
#undef feof
#undef ferror
#undef fflush
#undef fgetc
#undef fgets
#undef fgetpos
#undef fileno
#undef fopen
#undef fprintf
#undef fputs
#undef fputc
#undef fread
#undef fseek
#undef fsetpos
#undef fstat
#undef ftell
#undef fwrite
#undef getc
#undef putc
#undef rewind
#undef ungetc
#undef setbuf
#undef setvbuf
#undef tmpfile
#undef vfprintf
#undef popen
#undef _popen
#undef pclose
#undef _pclose
#undef pipe
#undef _pipe

#undef access
#undef chmod
#undef chsize
#undef close
#undef dup
#undef dup2
#undef isatty
#undef lseek
#undef mktemp
#undef open
#undef read
#undef rename
#undef setmode
#undef tmpnam
#undef umask
#undef unlink
#undef utime
#undef write

#undef execv
#undef execvp
#undef _putenv
#undef getenv

#undef mkdir
#undef chdir
#undef rmdir
#undef stat
#undef strerror

#undef perl_free
#undef perl_construct
#undef perl_parse
#undef perl_run
#undef perl_destruct

// #define DEBUG_TIMING

//CPerlExe

class CPerlExe : public IPerlSupport
{
	// memory allocation routines
	void *Malloc(size_t);
	void *Realloc(void*, size_t);
	void Free(void*);

	// FILE* routines
	FILE* Stdin(void);
	FILE* Stdout(void);
	FILE* Stderr(void);
	void Clearerr(FILE *);
	int Fclose(FILE *, int&);
	FILE *Fdopen(int, const char *, int&);
	int Feof(FILE *, int&);
	int Ferror(FILE *, int&);
	int FFlush(FILE *, int&);
	int Flushall(int&);
	int Fgetc(FILE *, int&);
	int FGetpos(FILE *, fpos_t *, int&);
	char *Fgets(char *, int, FILE *, int&);
	int Fileno(FILE *, int&);
	FILE *Fopen(const char *, const char *, int&);
	int Fputs(const char *, FILE *, int&);
	size_t Fread(void *, size_t, size_t, FILE *, int&);
	int Fseek(FILE *, long, int, int&);
	int FSetpos(FILE *, fpos_t *, int&);
	int FStat(int, struct stat *, int&);
	long Ftell(FILE *, int&);
	size_t Fwrite(const void *, size_t, size_t, FILE *, int&);
	int Getc(FILE *, int&);
	int Putc(int c, FILE *, int&);
	void Rewind(FILE *);
	void Setbuf(FILE *, char *, int&);
	int Setvbuf(FILE *, char *, int, size_t, int&);
	FILE* TmpFile(int&);
	int Ungetc(int, FILE *, int&);
	int Vfprintf(FILE *, const char *, va_list, int&);
	FILE *Popen(const char *, const char *, int&);
	int Pclose(FILE *, int&);
	int Pipe(int *, unsigned int, int, int&);

	// Low level I/O
	int Access(const char *, int, int&);
	int Chmod(const char *, int, int&);
	int Chsize(int, long, int&);
	int Close(int, int&);
	int Dup(int, int&);
	int Dup2(int, int, int&);
	int Isatty(int, int&);
	long Lseek(int, long, int, int&);
	char *Mktemp(char *, int&);
	int Open(const char *, int, int, int&);
	int Read(int, void *, unsigned int, int&);
	int Rename(const char *, const char *, int&);
	int Setmode(int, int, int&);
	char *Tmpnam(char *, int&);
	int Umask(int, int&);
	int Unlink(const char *, int&);
	int Utime(const char *filename, struct utimbuf *times, int&);
	int Write(int, const void *, unsigned int, int&);

	// Spawning
	int Spawn(int mode, const char *cmdname, const char * const *argv, const char * const *envp, int& e);
	int Cmd(int mode, const char *shell, const char *cmd, const char * const *envp);
	int Execv(const char *cmdname, const char *const *argv);
	int Execvp(const char *cmdname, const char *const *argv);

	// Sockets
	virtual int OpenOSfhandle(long osfhandle, int flags);
	virtual int GetOSfhandle(int filenum);

	// Misc
	int SetSignal(int i);
	void SignalHandler(CPerl *pPerl, int sig);
	void Die(CPerl *pPerl, char *message);
	void Putenv(const char *envstring);
	char* Getenv(const char *envstring);
	int GetCWD(char *buffer, int size);
	int SetCWD(char *buffer);
	void PerlCaller(char *string, int size);
	void* GetECB(void) { return NULL; };

	int MkDir(char *, int&);
	int ChDir(char *, int&);
	int RmDir(char *, int&);
	int FileStat(char *, void *, int&);
	char *GetStrerror(int);

	void XInit(void);

	// find file functions
	HANDLE FindFirstFile(const char *, void*);
	int FindNextFile(HANDLE, void *);
	int FindClose(HANDLE);

	void AbortPerl(void);

public:
	CPerlExe(void)
	{
		pDll = NULL;
		pPerl = NULL;
		bSignaled = FALSE;
		bSignalSet = FALSE;
	};

	inline BOOL PerlCreate(void)
	{
		try
		{
			pPerl = CreateCPerl(this, &pDll);
			if(pPerl != NULL)
			{
				try
				{
					pPerl->perl_construct();
				}
				catch(...)
				{
					fprintf(stderr, "%s\n", "Error: Unable to construct data structures");
					pPerl->perl_free();
					pPerl = NULL;
				}
			}
		}
		catch(...)
		{
			fprintf(stderr, "%s\n", "Error: Unable to allocate memory");
			pPerl = NULL;
		}
		return (pPerl != NULL);
	};
	inline int PerlParse(int argc, char** argv, char** env)
	{
		char* environ = NULL;
		int retVal;
		try
		{
			retVal = pPerl->perl_parse(NULL, argc, argv, (env == NULL || *env == NULL ? &environ : env));
		}
		catch(int x)
		{
			// this is where exit() should arrive
			retVal = x;
		}
		catch(...)
		{
			fprintf(stderr, "Error: Parse exception\n");
			retVal = -1;
		}
		return retVal;
	};
	inline int PerlRun(void)
	{
		int retVal;
		try
		{
			retVal = pPerl->perl_run();
		}
		catch(int x)
		{
			// this is where exit() should arrive
			retVal = x;
		}
		catch(...)
		{
			fprintf(stderr, "Error: Runtime exception\n");
			retVal = -1;
		}
		return retVal;
	};
	inline void PerlDestroy(void)
	{
		try
		{
			pPerl->perl_destruct();
			pPerl->perl_free();
		}
		catch(...)
		{
		}
	};

	inline void Signal(int sig)
	{
		SignalHandler(pPerl, sig);
	};

	inline void SetSignal(void)
	{
		if(!bSignalSet)
		{
			bSignaled = TRUE;
			bSignalSet = TRUE;
		}
	};

	inline void HadSignal(void)
	{
		if(bSignaled)
			AbortPerl();
	};

protected:
	IPerlSupport *pDll;
	CPerl *pPerl;
	BOOL bSignaled, bSignalSet;
};

void* CPerlExe::Malloc(size_t a)
{
	HadSignal();
	return(malloc(a));
}

void* CPerlExe::Realloc(void* a, size_t b)
{
	HadSignal();
	return(realloc(a, b));
}

void CPerlExe::Free(void* a)
{
	HadSignal();
	free(a);
}

FILE* CPerlExe::Stdin(void)
{
	HadSignal();
	return (&_iob[0]);
}

FILE* CPerlExe::Stdout(void)
{
	HadSignal();
	return (&_iob[1]);
}

FILE* CPerlExe::Stderr(void)
{
	HadSignal();
	return (&_iob[2]);
}

void CPerlExe::Clearerr(FILE *fp)
{
	HadSignal();
	clearerr(fp);
}

int CPerlExe::Fclose(FILE *a, int& e)
{
	HadSignal();
	int ret = fclose(a);
	if(errno)
		e = errno;
	return ret;
}

FILE* CPerlExe::Fdopen(int a, const char * b, int& e)
{
	HadSignal();
	FILE* ret = fdopen(a, b);
	if(errno)
		e = errno;
	return ret;
}

int CPerlExe::Feof(FILE *a, int& e)
{
	HadSignal();
	int ret = feof(a);
	if(errno)
		e = errno;
	return ret;
}

int CPerlExe::Ferror(FILE *a, int& e)
{
	HadSignal();
	int ret = ferror(a);
	if(errno)
		e = errno;
	return ret;
}

int CPerlExe::FFlush(FILE *a, int& e)
{
	HadSignal();
	int ret = fflush(a);
	if(errno)
		e = errno;
	return ret;
}

int CPerlExe::Flushall(int& e)
{
	HadSignal();
	int ret = _flushall();
	if(errno)
		e = errno;
	return ret;
}

int CPerlExe::Fgetc(FILE *a, int& e)
{
	HadSignal();
	int ret = fgetc(a);
	if(errno)
		e = errno;
	return ret;
}

int CPerlExe::FGetpos(FILE *a, fpos_t *b, int& e)
{
	HadSignal();
	int ret = fgetpos(a, b);
	if(errno)
		e = errno;
	return ret;
}

char* CPerlExe::Fgets(char *a, int b, FILE *c, int& e)
{
	HadSignal();
	char* ret = fgets(a, b, c);
	if(errno)
		e = errno;
	return ret;
}

int CPerlExe::Fileno(FILE *a, int& e)
{
	HadSignal();
	int ret = fileno(a);
	if(errno)
		e = errno;
	return ret;
}

FILE* CPerlExe::Fopen(const char *a, const char *b, int& e)
{
	HadSignal();
	FILE* ret = NULL;
	if(*a != '\0')
	{
		ret = fopen(a, b);
		if(errno)
			e = errno;
	}
	else
		e = EINVAL;		
	return ret;
}

int CPerlExe::Fputs(const char *a, FILE *b, int& e)
{
	HadSignal();
	int ret = fputs(a, b);
	if(errno)
		e = errno;
	return ret;
}

size_t CPerlExe::Fread(void *a, size_t b, size_t c, FILE *d, int& e)
{
	HadSignal();
	size_t ret = fread(a, b , c, d);
	if(errno)
		e = errno;
	return ret;
}

int CPerlExe::Fseek(FILE *a, long b, int c, int& e)
{
	HadSignal();
	int ret = fseek(a, b, c);
	if(errno)
		e = errno;
	return ret;
}

int CPerlExe::FSetpos(FILE *a, fpos_t *b, int& e)
{
	HadSignal();
	int ret = fsetpos(a, b);
	if(errno)
		e = errno;
	return ret;
}

int CPerlExe::FStat(int handle, struct stat *buf, int& e)
{
	HadSignal();
	int ret = fstat(handle, buf);
	if(errno)
		e = errno;
	return ret;
}

long CPerlExe::Ftell(FILE *a, int& e)
{
	HadSignal();
	long ret = ftell(a);
	if(errno)
		e = errno;
	return ret;
}

size_t CPerlExe::Fwrite(const void *a, size_t b, size_t c, FILE *d, int& e)
{
	HadSignal();
	size_t ret = fwrite(a, b, c,d);
	if(errno)
		e = errno;
	return ret;
}

int CPerlExe::Getc(FILE *a, int& e)
{
	HadSignal();
	int ret = getc(a);
	if(errno)
		e = errno;
	return ret;
}

int CPerlExe::Putc(int a, FILE *b, int& e)
{
	HadSignal();
	int ret = putc(a, b);
	if(errno)
		e = errno;
	return ret;
}

void CPerlExe::Rewind(FILE *a)
{
	HadSignal();
	rewind(a);
}

void CPerlExe::Setbuf(FILE *a, char *b, int&)
{
	HadSignal();
	setbuf(a, b);
}

int CPerlExe::Setvbuf(FILE *a , char *b, int c, size_t d, int& e)
{
	HadSignal();
	return setvbuf(a, b, c, d);
}

FILE* CPerlExe::TmpFile(int&)
{
	return tmpfile();
}

int CPerlExe::Ungetc(int a, FILE *b, int& e)
{
	HadSignal();
	int ret = ungetc(a, b);
	if(errno)
		e = errno;
	return ret;
}

int CPerlExe::Vfprintf(FILE *a, const char * b, va_list c, int& e)
{
	HadSignal();
	int ret = vfprintf(a, b, c);
	if(errno)
		e = errno;
	return ret;
}

FILE* CPerlExe::Popen(const char * a, const char * b, int& e)
{
	HadSignal();
	FILE* ret = _popen(a, b);
	if(errno)
		e = errno;
	return ret;
}

int CPerlExe::Pclose(FILE *a, int& e)
{
	HadSignal();
	int ret = _pclose(a);
	if(errno)
		e = errno;
	return ret;
}

int CPerlExe::Pipe(int * a, unsigned int b, int c, int& e)
{
	HadSignal();
	int ret = _pipe(a,b,c);
	if(errno)
		e = errno;
	return ret;
}


	// Low level I/O
int CPerlExe::Access(const char *a, int b, int& e)
{
	HadSignal();
	int ret = access(a, b);
	if(ret)
		e = errno;

	return ret;
}

int CPerlExe::Chmod(const char *a, int b, int& e)
{
	HadSignal();
	int ret = chmod(a, b);
	if(ret)
		e = errno;

	return ret;
}

int CPerlExe::Chsize(int a, long b, int& e)
{
	HadSignal();
	int ret = chsize(a, b);
	if(ret)
		e = errno;

	return ret;
}

int CPerlExe::Close(int a, int& e)
{
	HadSignal();
	int ret = close(a);
	if(errno)
		e = errno;
	return ret;
}

int CPerlExe::Dup(int a, int& e)
{
	HadSignal();
	int ret = dup(a);
	if(errno)
		e = errno;
	return ret;
}

int CPerlExe::Dup2(int a, int b, int& e)
{
	HadSignal();
	int ret = dup2(a,b);
	if(errno)
		e = errno;
	return ret;
}

int CPerlExe::Isatty(int a, int& e)
{
	HadSignal();
	return isatty(a);
}

long CPerlExe::Lseek(int a, long b, int c, int& e)
{
	HadSignal();
	long ret = lseek(a,b,c);
	if(errno)
		e = errno;
	return ret;
}

char *CPerlExe::Mktemp(char *a , int& e)
{
	HadSignal();
	return mktemp(a);
}

int CPerlExe::Open(const char * a, int b, int c, int& e)
{
	HadSignal();
	int ret = open(a,b, c);
	if(errno)
		e = errno;
	return ret;
}

int CPerlExe::Read(int a, void * b, unsigned int c, int& e)
{
	HadSignal();
	int ret = read(a,b,c);
	if(errno)
		e = errno;
	return ret;
}

int CPerlExe::Rename(const char *OldFileName, const char *newname, int& e)
{
	HadSignal();
	char szNewWorkName[MAX_PATH+1];
	WIN32_FIND_DATA fdOldFile, fdNewFile;
	HANDLE handle;
	char *ptr;

	if((strchr(OldFileName, '\\') || strchr(OldFileName, '/'))
		&& strchr(newname, '\\') == NULL
			&& strchr(newname, '/') == NULL)
	{
		strcpy(szNewWorkName, OldFileName);
		if((ptr = strrchr(szNewWorkName, '\\')) == NULL)
			ptr = strrchr(szNewWorkName, '/');
		strcpy(++ptr, newname);
	}
	else
		strcpy(szNewWorkName, newname);

	if(stricmp(OldFileName, szNewWorkName) != 0)
	{	// check that we're not being fooled by relative paths
		// and only delete the new file if it exists and is not the same file as the old file
		// GetFullPathName does not return the long file name on some systems
		handle = FindFirstFile(OldFileName, &fdOldFile);
		if(handle != NULL)
			FindClose(handle);
		
		handle = FindFirstFile(szNewWorkName, &fdNewFile);
		if(handle != NULL)
			FindClose(handle);
		else
			fdNewFile.cFileName[0] = '\0';

		if(strcmp(fdOldFile.cAlternateFileName, fdNewFile.cAlternateFileName) != 0)
		{	// file exists and not same file
			DeleteFile(szNewWorkName);
		}
	}
	int ret = rename(OldFileName, szNewWorkName);
	if(ret)
		e = errno;

	return ret;
}

int CPerlExe::Setmode(int a, int b, int& e)
{
	HadSignal();
	int ret = setmode(a, b);
	if(ret)
		e = errno;

	return ret;
}

char *CPerlExe::Tmpnam(char *a, int& e)
{
	HadSignal();
	return tmpnam(a);
}

int CPerlExe::Umask(int a, int& e)
{
	HadSignal();
	return umask(a);
}

int CPerlExe::Unlink(const char *a, int& e)
{
	HadSignal();
	chmod(a, _S_IREAD | _S_IWRITE);
	int ret = unlink(a);
	if(ret)
		e = errno;

	return ret;
}

int CPerlExe::Utime(const char *filename, struct utimbuf *times, int& e)
{
	HadSignal();
	int ret = utime(filename, times);
	if(ret)
		e = errno;
	return ret;
}

int CPerlExe::Write(int a, const void * b, unsigned int c, int& e)
{
	HadSignal();
	int ret = write(a,b,c);
	if(errno)
		e = errno;
	return ret;
}

int CPerlExe::Spawn(int mode, const char *cmdname, const char * const *argv, const char * const *envp, int& e)
{
//	since _putenv is being called on this side
//	we can call Spawn with our environ
	HadSignal();
	int ret = spawnvpe(mode, cmdname, argv, (const char * const *)environ);
	if(ret == -1)
		e = errno;
	return ret;
}

int CPerlExe::Cmd(int mode, const char *shell, const char *cmd, const char * const *envp)
{
//	since _putenv is being called on this side
//	we can call Spawn with our environ
	HadSignal();
	return spawnle(mode, shell, shell, "/c", cmd, (char*)0, (const char * const *)environ);
}

int CPerlExe::Execv(const char *cmdname, const char *const *argv)
{
	HadSignal();
	return execv(cmdname, argv);
}

int CPerlExe::Execvp(const char *cmdname, const char *const *argv)
{
	HadSignal();
	return execvp(cmdname, argv);
}

#ifdef _X86_
extern "C" int __cdecl _alloc_osfhnd(void);
extern "C" int __cdecl _set_osfhnd(int fh, long value);
extern "C" void __cdecl _unlock(int);

#if (_MSC_VER >= 1000)
typedef struct
{
	long osfhnd;    /* underlying OS file HANDLE */
	char osfile;    /* attributes of file (e.g., open in text mode?) */
	char pipech;    /* one char buffer for handles opened on pipes */
}	ioinfo;
extern "C" ioinfo * __pioinfo[];
#define IOINFO_L2E			5
#define IOINFO_ARRAY_ELTS	(1 << IOINFO_L2E)
#define _pioinfo(i)	(__pioinfo[i >> IOINFO_L2E] + (i & (IOINFO_ARRAY_ELTS - 1)))
#define _osfile(i)	(_pioinfo(i)->osfile)
#else
extern "C" extern char _osfile[];
#endif	// (_MSC_VER >= 1000)

#define FOPEN			0x01	// file handle open
#define FAPPEND			0x20	// file handle opened O_APPEND
#define FDEV			0x40	// file handle refers to device
#define FTEXT			0x80	// file handle is in text mode

#define _STREAM_LOCKS   26		// Table of stream locks
#define _LAST_STREAM_LOCK  (_STREAM_LOCKS+_NSTREAM_-1)	// Last stream lock
#define _FH_LOCKS          (_LAST_STREAM_LOCK+1)		// Table of fh locks
#endif	// _X86_

int CPerlExe::OpenOSfhandle(long osfhandle, int flags)
{
#ifdef _X86_
	int fh;
	char fileflags;		// _osfile flags 

	// copy relevant flags from second parameter 
	fileflags = FDEV;

	if(flags & _O_APPEND)
		fileflags |= FAPPEND;

	if(flags & _O_TEXT)
		fileflags |= FTEXT;

	// attempt to allocate a C Runtime file handle
	if((fh = _alloc_osfhnd()) == -1)
	{
		errno = EMFILE;		// too many open files 
		_doserrno = 0L;		// not an OS error
		return -1;			// return error to caller
	}

	// the file is open. now, set the info in _osfhnd array
	_set_osfhnd(fh, osfhandle);

	fileflags |= FOPEN;			// mark as open

#if (_MSC_VER >= 1000)
	_osfile(fh) = fileflags;	// set osfile entry
#else
	_osfile[fh] = fileflags;	// set osfile entry
#endif

	return fh;					// return handle
#else
	return _open_osfhandle(osfhandle, flags);
#endif	// _X86_
}

int CPerlExe::GetOSfhandle(int filenum)
{
	return _get_osfhandle(filenum);
}

CPerlExe *g_sig;
void sigHandler(int sig)
{
	if(g_sig != NULL)
		g_sig->Signal(sig);

	else if(sig == SIGINT)
		exit(sig);
}

int CPerlExe::SetSignal(int i)
{
	HadSignal();
	g_sig = this;
	signal(i, sigHandler);
	return 1;
}

void CPerlExe::SignalHandler(CPerl *pPerl, int sig)
{
	HadSignal();
	pDll->SignalHandler(pPerl, sig);
}

void CPerlExe::Die(CPerl *pPerl, char *message)
{
	HadSignal();
	pDll->Die(pPerl, message);
}

void CPerlExe::Putenv(const char *envstring)
{
	HadSignal();
	_putenv(envstring);
}

char* CPerlExe::Getenv(const char *envstring)
{
	HadSignal();
	return getenv(envstring);
}

int CPerlExe::GetCWD(char *buffer, int size)
{
	HadSignal();
	return GetCurrentDirectory(size, buffer);
}

int CPerlExe::SetCWD(char *buffer)
{
	HadSignal();
	return SetCurrentDirectory(buffer);
}

void CPerlExe::PerlCaller(char *string, int size)
{
	char *ptr = "PerlExe";

	while(--size && (*string++ = *ptr++) != '\0')
		;

	*string = '\0';
}

int CPerlExe::MkDir(char *pathName, int& e)
{
	int ret = mkdir(pathName);
	if(ret == -1)
		e = errno;

	return ret;
}

int CPerlExe::ChDir(char *pathName, int& e)
{
	int ret = chdir(pathName);
	if(ret == -1)
		e = errno;

	return ret;
}

int CPerlExe::RmDir(char *pathName, int& e)
{
	int ret = rmdir(pathName);
	if(ret == -1)
		e = errno;

	return ret;
}

int CPerlExe::FileStat(char *fileName, void *buffer, int& e)
{
	int ret = stat(fileName, (struct stat*)buffer);
	if(ret == -1)
		e = errno;

	return ret;
}

char *CPerlExe::GetStrerror(int e)
{
//
// Since the errors returned by the socket error function 
// WSAGetLastError() are not known by the library routine strerror
// we have to roll our own.
//
	static char buffer[512];
	extern int sys_nerr;
	DWORD source = 0;

	if(e < 0 || e > sys_nerr) 
	{
		if(e < 0)
			e = GetLastError();

		if(FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM, &source, e, 0, buffer, sizeof(buffer), NULL) == 0) 
			sprintf(buffer, "Unknown Error: 0x%08lx", e);

		return buffer;
	}

	return strerror(e);
}

extern void XInit(void* pVoid);

void CPerlExe::XInit(void)
{
	::XInit(pPerl);
}

HANDLE CPerlExe::FindFirstFile(const char *pStr, void*pData)
{
	return ::FindFirstFile(pStr, (LPWIN32_FIND_DATA)pData);
}

int CPerlExe::FindNextFile(HANDLE hHandle, void *pData)
{
	return ::FindNextFile(hHandle, (LPWIN32_FIND_DATA)pData);
}

int CPerlExe::FindClose(HANDLE hHandle)
{
	return ::FindClose(hHandle);
}

void CPerlExe::AbortPerl(void)
{
	bSignaled = FALSE;
	pDll->Die(pPerl, "^C");
}

CPerlExe *g_sigInt;
void sigIntHandler(int sig)
{
	g_sigInt->SetSignal();
}



void main(int argc, char **argv, char **env)
{
	int exitstatus = 1;
	CPerlExe exe;

	if(PerlGetVersion() < 300)
	{
#ifdef PRIVATE_PERL
		fprintf(stderr, "Perl.exe requires pPerl300.dll build 300 or later");
#else
		fprintf(stderr, "Perl.exe requires perl300.dll build 300 or later");
#endif
		exit(1);
	}

#ifdef DEBUG_TIMING
	// For timing execution
	FILE *fp;
	DWORD dwStart, dwAfterCompile, dwEnd;

	dwStart = GetTickCount();
#endif

	if(exe.PerlCreate())
	{
		exitstatus = exe.PerlParse(argc, argv, env);
		if(exitstatus == 0)
		{
			g_sigInt = &exe;
			signal(SIGINT, sigIntHandler);

#ifdef DEBUG_TIMING
			dwAfterCompile = GetTickCount();
#endif
			exitstatus = exe.PerlRun();

		}

		exe.PerlDestroy();
	
#ifdef DEBUG_TIMING
		dwEnd = GetTickCount();

		fp = fopen("Timing.log", "at");
		if(fp != NULL)
		{	// output tab delim line with script name, start time, after compile time, end time, and newline
			fprintf(fp, "%s\t%lu\t%lu\t%lu\n", argv[argc-1], dwStart, dwAfterCompile, dwEnd);
			fclose(fp);
		}
#endif

	}

	exit(exitstatus);
}

