//
// XInit.cpp
//
//

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include "EXTERN.h"
#include "perl.h"
#include "XSub.h"

#define CROAK croak
#define PERLADVAPI WINAPI

// Functions from User32.Dll
typedef int (PERLADVAPI *LPMessageBox)(HWND, LPCTSTR, LPCTSTR, UINT);
typedef HWND (PERLADVAPI *LPGetActiveWindow)(void);

// Functions from User32.Dll
static LPMessageBox pMessageBox = 0;
static LPGetActiveWindow pGetActiveWindow = 0;

static HINSTANCE hUserApiLib = NULL;


static FARPROC GetUserApiAddress(CPerl *pPerl, HINSTANCE hInstance, char *lpFunctionName)
{
	char buffer[512];
	FARPROC proc = GetProcAddress(hInstance, lpFunctionName);
	if(proc == 0)
	{
		sprintf(buffer, "Unable to get address of %s in USER32.DLL", lpFunctionName);
		CROAK(buffer);
	}
	return proc;
}


void UserApiStartup(CPerl *pPerl)
{
	if(hUserApiLib == NULL)
	{
		HINSTANCE hLib = LoadLibrary("USER32.DLL");
		if(hLib == NULL)
			CROAK("Could not load USER32.dll\n");

		pMessageBox = (LPMessageBox)GetUserApiAddress(pPerl, hLib, "MessageBoxA");
		pGetActiveWindow = (LPGetActiveWindow)GetUserApiAddress(pPerl, hLib, "GetActiveWindow");
		hUserApiLib = hLib;
	}
}

void __cdecl NTADVAPICleanUp(void)
{
	if(hUserApiLib != NULL)
	{
		FreeLibrary(hUserApiLib);
		hUserApiLib = NULL;
	}
}

#define USERSTARTUP	  if(hUserApiLib == 0) UserApiStartup(pPerl)

XS(NTMsgBox)
{
	dXSARGS;
	char *msg, *title;
	DWORD dwFlag = MB_ICONEXCLAMATION;

	if(items != 1 && items != 2 && items != 3)
	{
		CROAK("usage: Win32::MsgBox($Message [, $Buttons [, $Title]]);\n");
	}

	USERSTARTUP;
	title = "Perl for Win32";
	msg = SvPV(ST(0), na);
	if(items > 1)
	{
		dwFlag = SvIV(ST(1));
		if(items > 2)
			title = SvPV(ST(2), na);
	}
	XSRETURN_IV(pMessageBox(pGetActiveWindow(), msg, title, dwFlag));
}

XS(NTInputBox)
{
//	dSP;
//	PUSHMARK(sp);
//	PUTBACK;
}

void XInit(void* pVoid)
{
	static char thisFile[] = __FILE__;
	CPerl *pPerl = (CPerl*)pVoid;

	newXS("Win32::MsgBox", NTMsgBox, thisFile);
	newXS("Win32::InputBox", NTInputBox, thisFile);
}


